"""header base class."""
from __future__ import annotations

import abc
import typing

HeaderT = typing.TypeVar("HeaderT", bound="Header")


class Header(abc.ABC):
    """Abstract base class for a message header."""

    length = -1

    def __init__(  # pylint: disable=too-many-arguments
        self,
        system: int,
        session_id: int,
        stream: int,
        function: int,
        require_response: bool
    ):
        """Initialize a hsms header.

        Args:
            system: message ID
            session_id: device / session ID
            stream: stream
            function: function
            require_response: True when a response is required

        """
        self._session_id = session_id
        self._stream = stream
        self._function = function
        self._system = system
        self._require_response = require_response


    @property
    def session_id(self) -> int:
        """Get session id."""
        return self._session_id

    @property
    def stream(self) -> int:
        """Get stream."""
        return self._stream

    @property
    def function(self) -> int:
        """Get function."""
        return self._function

    @property
    def system(self) -> int:
        """Get system."""
        return self._system

    @property
    def require_response(self) -> bool:
        """Get require response flag."""
        return self._require_response

    @abc.abstractmethod
    def encode(self) -> bytes:
        """Encode header to message.

        Returns:
            encoded header

        """
        raise NotImplementedError("Header.encode missing implementation")

    @classmethod
    @abc.abstractmethod
    def decode(cls, data: bytes) -> Header:
        """Decode byte array header to Header object.

        Args:
            data: byte-encode header data

        Returns:
            Header object

        """
        raise NotImplementedError("Header.decode missing implementation")

    @property
    @abc.abstractmethod
    def _as_dictionary(self) -> dict[str, typing.Any]:
        """Get the data as dictionary.

        Returns:
            Header data as dictionary.

        """
        raise NotImplementedError("Header._as_dictionary missing implementation")

    def updated_with(self: HeaderT, **kwargs) -> HeaderT:
        """Get a new header with updated fields.

        Args:
            kwargs: parameter name will update constructor field

        Returns:
            new header with modified data

        """
        data = self._as_dictionary

        data.update(kwargs)

        return self.__class__(**data)



"""Class for stream 02 function 46."""

from secsgem.secs.data_items import LIMITACK, LIMITID, LVACK, VID, VLAACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F46(SecsStreamFunction):
    """Define variable limit attributes - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F46
        {
            VLAACK: B[1]
            DATA: [
                {
                    VID: U1/U2/U4/U8/I1/I2/I4/I8/A
                    LVACK: B[1]
                    DATA: {
                        LIMITID: B[1]
                        LIMITACK: B[1]
                    }
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F46({
        ...     "VLAACK": secsgem.secs.data_items.VLAACK.LIMIT_DEF_ERROR,
        ...     "DATA": [{
        ...         "VID": 2,
        ...         "LVACK": secsgem.secs.data_items.LVACK.VARIABLE_UNKNOWN,
        ...         "DATA": {
        ...             "LIMITID": 3,
        ...             "LIMITACK": secsgem.secs.data_items.LIMITACK.ASCII_ILLEGAL
        ...         }
        ...     }]})
        S2F46
          <L [2]
            <B 0x1>
            <L [1]
              <L [3]
                <U1 2 >
                <B 0x1>
                <L [2]
                  <B 0x3>
                  <B 0x6>
                >
              >
            >
          > .

    Data Items:
        - :class:`VLAACK <secsgem.secs.data_items.VLAACK>`
        - :class:`VID <secsgem.secs.data_items.VID>`
        - :class:`LVACK <secsgem.secs.data_items.LVACK>`
        - :class:`LIMITID <secsgem.secs.data_items.LIMITID>`
        - :class:`LIMITACK <secsgem.secs.data_items.LIMITACK>`

    """

    _stream = 2
    _function = 46

    _data_format = [
        VLAACK,
        [
            [
                VID,
                LVACK,
                [
                    LIMITID,
                    LIMITACK
                ]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""TCP server connection."""
from __future__ import annotations

import select
import socket
import threading
import time
import typing

from .helpers import is_windows
from .tcp_connection import TcpConnection

if typing.TYPE_CHECKING:
    from .settings import Settings


class TcpServerConnection(TcpConnection):
    """Server class for TCP server connection.

    Creates a listening socket and waits for one incoming connection on this socket.
    After the connection is established the listening socket is closed.
    """

    def __init__(self, settings: Settings):
        """Initialize a TCP server connection.

        Args:
            settings: protocol and communication settings

        """
        # initialize super class
        TcpConnection.__init__(self, settings)

        # initially not enabled
        self._enabled = False

        # reconnect thread required for server
        self._server_thread = None
        self._stop_server_thread = False
        self._server_sock = None

        self.on_disconnected.register(self._disconnected)

    def _disconnected(self, _: dict[str, typing.Any]):
        """Called when the connection was disconnected.

        This is required to initiate the reconnect if the connection is still enabled

        """
        if self._enabled:
            self.__start_server_thread()

    def enable(self):
        """Enable the connection.

        Starts the connection server process.
        """
        # only start if not already enabled
        if not self._enabled:
            # mark connection as enabled
            self._enabled = True

            # start the connection thread
            self.__start_server_thread()

    def disable(self):
        """Disable the connection.

        Stops all connection attempts, and closes the connection
        """
        # only stop if enabled
        if self._enabled:
            # mark connection as disabled
            self._enabled = False

            # stop connection thread if it is running
            if self._server_thread and self._server_thread.is_alive():
                self._stop_server_thread = True

                if self._server_sock:
                    self._server_sock.close()

                # wait for connection thread to stop
                while self._stop_server_thread:
                    time.sleep(0.2)

            # disconnect super class
            self.disconnect()

    def __start_server_thread(self):
        self._server_thread = threading.Thread(
            target=self.__server_thread,
            name=f"secsgem_tcpServerConnection_serverThread_{self._settings.address}")
        self._server_thread.start()

    def __server_thread(self):
        """Thread function to wait for incoming tcp connections.

        .. warning:: Do not call this directly, for internal use only.
        """
        self._server_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if not is_windows():
            self._server_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        self._server_sock.bind((self._settings.address, self._settings.port))
        self._server_sock.listen(1)

        while not self._stop_server_thread:
            try:
                select_result = select.select([self._server_sock], [], [], self.select_timeout)
                if not select_result[0]:
                    # select timed out
                    continue
            except Exception as exc:  # pylint: disable=broad-except
                self._logger.debug("select exception", exc_info=exc)

            accept_result = self._server_sock.accept()
            if accept_result is None:
                continue

            (self._sock, (_, _)) = accept_result

            # setup socket
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

            # make socket nonblocking
            self._socket.setblocking(False)

            # mark connection as connected
            self._connected = True

            # start the receiver thread
            self._start_receiver()

            # send event
            try:
                self.on_connected({"source": self})
                self._server_sock.shutdown(socket.SHUT_RDWR)
                self._server_sock.close()
            except OSError:  # pylint: disable=broad-except
                self._logger.warning("Ignoring exception for on_connection_established handler.")

            return

        self._stop_server_thread = False



"""protocol base class."""
from __future__ import annotations

import abc
import logging
import queue
import random
import typing

from .block_send_info import BlockSendInfo
from .byte_queue import ByteQueue
from .events import EventProducer
from .protocol_dispatcher import ProtocolDispatcher

if typing.TYPE_CHECKING:
    from ..secs.functions.base import SecsStreamFunction
    from .connection import Connection
    from .message import Message
    from .settings import Settings

MessageT = typing.TypeVar("MessageT", bound="Message")
BlockT = typing.TypeVar("BlockT", bound="Block")


class Protocol(abc.ABC, typing.Generic[MessageT, BlockT]):  # pylint: disable=too-many-instance-attributes
    """Abstract base class for a protocol."""

    message_type: type[MessageT]

    def __init__(self, settings: Settings) -> None:
        """Initialize protocol base object."""
        super().__init__()

        self._settings = settings

        self._event_producer = EventProducer()
        self._event_producer.targets += self

        self._system_counter = random.randint(0, (2 ** 32) - 1)  # noqa: S311

        self._logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        self._communication_logger = logging.getLogger("communication")

        self.__connection: Connection | None = None

        self._response_queues: dict[int, queue.Queue[MessageT]] = {}

        self._receive_buffer = ByteQueue()
        self._send_queue: queue.Queue[BlockSendInfo] = queue.Queue()
        self._incomplete_messages: dict[int, MessageT] = {}

        self._thread = ProtocolDispatcher(
            self._process_data,
            self._dispatch_block,
            self._settings,
        )

    @property
    def communication_logger(self):
        return self._communication_logger

    @property
    def _connection(self) -> Connection:
        if self.__connection is None:
            self.__connection = self._settings.create_connection()
            self.__connection.on_connected.register(self._on_connected)
            self.__connection.on_data.register(self._on_connection_data_received)
            self.__connection.on_disconnecting.register(self._on_disconnecting)
            self.__connection.on_disconnected.register(self._on_disconnected)

        return self.__connection

    @abc.abstractmethod
    def _on_connected(self, _: dict[str, typing.Any]):
        raise NotImplementedError("Protocol._on_connected missing implementation")

    @abc.abstractmethod
    def _on_disconnecting(self, _: dict[str, typing.Any]):
        raise NotImplementedError("Protocol._on_disconnecting missing implementation")

    @abc.abstractmethod
    def _on_disconnected(self, _: dict[str, typing.Any]):
        raise NotImplementedError("Protocol._on_disconnected missing implementation")

    def _on_connection_data_received(self, data: dict[str, typing.Any]):
        """Data received by connection.

        Args:
            data: received data

        """
        self._receive_buffer.append(data["data"])
        self._thread.trigger_receiver()

    def _process_data(self):
        """Parse the receiving buffer and dispatch callbacks."""
        self._process_send_queue()
        self._process_received_data()

    @abc.abstractmethod
    def _process_send_queue(self):
        """Process the send to communication queue."""
        raise NotImplementedError("Protocol._process_send_queue missing implementation")

    @abc.abstractmethod
    def _process_received_data(self):
        """Process the receiving from communication queue."""
        raise NotImplementedError("Protocol._process_received_data missing implementation")

    def _dispatch_block(self, source: object, block: BlockT):
        result = self._add_message_block(block)
        if result is None:
            return

        try:
            self._on_connection_message_received(source, result)
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_connection_message_received handler")

    @property
    def events(self):
        """Property for event handling."""
        return self._event_producer

    def get_next_system_counter(self):
        """Return the next System.

        Returns:
            System for the next command

        """
        self._system_counter += 1

        if self._system_counter > ((2 ** 32) - 1):
            self._system_counter = 0

        return self._system_counter

    def enable(self):
        """Enable the connection."""
        self._connection.enable()

    def disable(self):
        """Disable the connection."""
        self._connection.disable()

    @abc.abstractmethod
    def serialize_data(self) -> dict[str, typing.Any]:
        """Get protocol serialized data for debugging."""
        raise NotImplementedError("Protocol.serialize_data missing implementation")

    @abc.abstractmethod
    def _on_connection_message_received(self, source: object, message: MessageT):
        """Message received by connection.

        Args:
            source: source of event
            message: received data message

        """
        raise NotImplementedError("Protocol._on_connection_message_received missing implementation")

    def _get_queue_for_system(self, system_id: int) -> queue.Queue:
        """Create a new queue to receive responses for a certain system.

        Args:
            system_id: system id to watch

        Returns:
            queue to receive responses with

        """
        self._response_queues[system_id] = queue.Queue()
        return self._response_queues[system_id]

    def _remove_queue(self, system_id: int):
        """Remove queue for system id from list.

        Args:
            system_id: system id to remove

        """
        del self._response_queues[system_id]

    def _add_message_block(self, block: BlockT) -> MessageT | None:
        """Add a block, and get completed message if available.

        Args:
            block: block to add

        Returns:
            completed message or None if paket not complete

        """
        if block.header.system not in self._incomplete_messages:
            self._incomplete_messages[block.header.system] = self.message_type.from_block(block)
        else:
            self._incomplete_messages[block.header.system].blocks.append(block)

        message = self._incomplete_messages[block.header.system]

        if not message.complete:
            return None

        del self._incomplete_messages[block.header.system]
        return message

    @abc.abstractmethod
    def _create_message_for_function(
            self,
            function: SecsStreamFunction,
            system_id: int,
    ) -> Message:
        """Create a protocol specific message for a function.

        Args:
            function: function to create message for
            system_id: system

        Returns:
            created message

        """
        raise NotImplementedError("Protocol._create_message_for_function missing implementation")

    def send_message(self, message: Message) -> bool:
        """Send a message to the remote host.

        Args:
            message: message to be transmitted

        Returns:
            True if sending was successful

        """
        for block in message.blocks:
            block_send_info = BlockSendInfo(block.encode())
            self._send_queue.put(block_send_info)
            self._thread.trigger_receiver()

            if not block_send_info.wait():
                return False

        return True

    def send_and_waitfor_response(self, function: SecsStreamFunction) -> Message | None:
        """Send the message and wait for the response.

        Args:
            function: message to be sent

        Returns:
            Message that was received

        """
        system_id = self.get_next_system_counter()

        response_queue = self._get_queue_for_system(system_id)

        out_message = self._create_message_for_function(function, system_id)

        self._communication_logger.info("> %s\n%s", out_message, function, extra=self._get_log_extra())

        if not self.send_message(out_message):
            self._logger.error("Sending message failed")
            self._remove_queue(system_id)
            return None

        try:
            response = response_queue.get(True, self._settings.timeouts.t3)
        except queue.Empty:
            response = None

        self._remove_queue(system_id)

        return response

    def send_response(self, function: SecsStreamFunction, system: int) -> bool:
        """Send response function for system.

        Args:
            function: function to be sent
            system: system to reply to

        Returns:
            True if sending was successful

        """
        out_message = self._create_message_for_function(function, system)

        self._communication_logger.info("> %s\n%s", out_message, function, extra=self._get_log_extra())

        return self.send_message(out_message)

    def send_stream_function(self, function: SecsStreamFunction) -> bool:
        """Send the message and wait for the response.

        Args:
            function: message to be sent

        Returns:
            True if successful

        """
        out_message = self._create_message_for_function(function, self.get_next_system_counter())

        self._communication_logger.info("> %s\n%s", out_message, function, extra=self._get_log_extra())

        return self.send_message(out_message)

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        return f"{self.__class__.__name__} {self.serialize_data()}"

    @abc.abstractmethod
    def _get_log_extra(self) -> dict[str, typing.Any]:
        """Get extra fields for logging."""
        raise NotImplementedError("Protocol._get_log_extra missing implementation")



"""Class for stream 01 function 04."""

from secsgem.secs.data_items import SV
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F04(SecsStreamFunction):
    """selected equipment status - data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F04
        [
            SV: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F04([secsgem.secs.variables.U1(1), "text", secsgem.secs.variables.U4(1337)])
        S1F4
          <L [3]
            <U1 1 >
            <A "text">
            <U4 1337 >
          > .

    Data Items:
        - :class:`SV <secsgem.secs.data_items.SV>`

    """

    _stream = 1
    _function = 4

    _data_format = [SV]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""TID data item."""
from .. import variables
from .base import DataItemBase


class TID(DataItemBase):
    """Terminal ID.

    :Type: :class:`Binary <secsgem.secs.variables.Binary>`
    :Length: 1

    **Used In Function**
        - :class:`SecsS10F01 <secsgem.secs.functions.SecsS10F01>`
        - :class:`SecsS10F03 <secsgem.secs.functions.SecsS10F03>`

    """

    __type__ = variables.Binary
    __count__ = 1



"""Class for stream 07 function 02."""

from secsgem.secs.data_items import PPGNT
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F02(SecsStreamFunction):
    """process program load - grant.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F02
        PPGNT: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F02(secsgem.secs.data_items.PPGNT.OK)
        S7F2
          <B 0x0> .

    Data Items:
        - :class:`PPGNT <secsgem.secs.data_items.PPGNT>`

    """

    _stream = 7
    _function = 2

    _data_format = PPGNT

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Handler for GEM commands."""
from __future__ import annotations

import logging
import threading
import typing

import secsgem.common
import secsgem.secs

from .communication_state_machine import CommunicationState, CommunicationStateMachine


class GemHandler(secsgem.secs.SecsHandler):  # pylint: disable=too-many-instance-attributes
    """Baseclass for creating Host/Equipment models. This layer contains GEM functionality."""

    def __init__(self, settings: secsgem.common.Settings):
        """Initialize a gem handler.

        Inherit from this class and override required functions.

        Args:
            settings: communication settings

        """
        super().__init__(settings)
        self._protocol.events.communicating += self._on_communicating

        self._mdln = "secsgem"  #: model number returned by S01E13/14
        self._softrev = "0.1.0"  #: software version returned by S01E13/14

        self._logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)

        self._is_host = True

        self._communication_state = CommunicationStateMachine(self.settings)
        self._communication_state.wait_cra.events.enter.register(self._on_state_wait_cra)
        self._communication_state.communicating.events.enter.register(self._on_state_communicating)

        self._report_id_counter = 1000

        self._wait_event_list: list[threading.Event] = []

    def __repr__(self) -> str:
        """Generate textual representation for an object of this class."""
        return f"{self.__class__.__name__} {self.serialize_data()}"

    @property
    def model_name(self):
        return self._mdln

    @model_name.setter
    def model_name(self, model_name):
        self._mdln = model_name

    @property
    def software_version(self):
        return self._softrev

    @software_version.setter
    def software_version(self, software_version):
        self._softrev = software_version

    @property
    def communication_state(self) -> CommunicationStateMachine:
        """Get the communication state model."""
        return self._communication_state

    def serialize_data(self) -> dict[str, typing.Any]:
        """Get serialized data.

        Returns:
            data to serialize for this object

        """
        data = self.protocol.serialize_data()
        data.update({"communicationState": self._communication_state.current,
                     "commDelayTimeout": self.settings.establish_communication_timeout,
                     "reportIDCounter": self._report_id_counter})
        return data

    def enable(self) -> None:
        """Enable the connection."""
        self._communication_state.enable()
        self.protocol.enable()

        self._logger.info("Connection enabled")

    def disable(self) -> None:
        """Disable the connection."""
        self.protocol.disable()
        self._communication_state.disable()

        self._logger.info("Connection disabled")

    def _on_message_received(self, data: dict[str, typing.Any]):
        """Message received from protocol layer.

        Args:
            data: received event data

        """
        message = data["message"]
        if self._communication_state.current == CommunicationState.WAIT_CRA:
            if message.header.stream == 1 and message.header.function == 13:
                if self._is_host:
                    self.send_response(self.stream_function(1, 14)({"COMMACK": self.on_commack_requested(),
                                                                    "MDLN": []}),
                                       message.header.system)
                else:
                    self.send_response(self.stream_function(1, 14)({"COMMACK": self.on_commack_requested(),
                                                                    "MDLN": [self._mdln, self._softrev]}),
                                       message.header.system)

                self._communication_state.s1f13received()
            elif message.header.stream == 1 and message.header.function == 14:
                self._communication_state.s1f14received()
        elif self._communication_state.current == CommunicationState.WAIT_DELAY:
            pass
        elif self._communication_state.current == CommunicationState.COMMUNICATING:
            self._handle_stream_function(message)

    def _on_communicating(self, _):
        """Selected received from hsms layer."""
        self._communication_state.select()

    def _on_state_wait_cra(self, _):
        """Connection state model changed to state WAIT_CRA.

        Args:
            data: event attributes

        """
        if self._is_host:
            self.send_stream_function(self.stream_function(1, 13)())
        else:
            self.send_stream_function(self.stream_function(1, 13)([self._mdln, self._softrev]))

    def _on_state_communicating(self, _):
        """Connection state model changed to state COMMUNICATING.

        Args:
            data: event attributes

        """
        self.events.fire("handler_communicating", {"handler": self})

        for event in self._wait_event_list:
            event.set()

    def on_connection_closed(self, _):
        """Handle connection was closed event."""
        self._logger.info("Connection was closed")

        if self._communication_state.current == CommunicationState.COMMUNICATING:
            # update communication state
            self._communication_state.communicationfail()

    def on_commack_requested(self) -> int:
        """Get the acknowledgement code for the connection request.

        override to accept or deny connection request

        Returns:
            0 when connection is accepted, 1 when connection is denied

        """
        return 0

    def send_process_program(self,
                             ppid: int | str,
                             ppbody: str):
        """Send a process program.

        Args:
            ppid: Transferred process programs ID
            ppbody: Content of process program

        """
        # send remote command
        self._logger.info("Send process program %s", ppid)

        return self.settings.streams_functions.decode(self.send_and_waitfor_response(self.stream_function(7, 3)(
            {"PPID": ppid, "PPBODY": ppbody}))).get()

    def request_process_program(self,
                                ppid: int | str) -> tuple[int | str, str]:
        """Request a process program.

        ppid: Transferred process programs ID

        """
        self._logger.info("Request process program %s", ppid)

        # send remote command
        s7f6 = self.settings.streams_functions.decode(self.send_and_waitfor_response(self.stream_function(7, 5)(ppid)))
        return s7f6.PPID.get(), s7f6.PPBODY.get()

    def waitfor_communicating(self, timeout: float | None = None) -> bool:
        """Wait until connection gets into communicating state. Returns immediately if state is communicating.

        Args:
            timeout: seconds to wait before aborting

        Returns:
            True if state is communicating, False if timed out

        """
        event = threading.Event()
        self._wait_event_list.append(event)

        if self._communication_state.current == CommunicationState.COMMUNICATING:
            self._wait_event_list.remove(event)
            return True

        result = event.wait(timeout)

        self._wait_event_list.remove(event)

        return result

    def _on_s01f01(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 1, Function 1, Are You There.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler, message  # unused parameters

        if self._is_host:
            return self.stream_function(1, 2)()

        return self.stream_function(1, 2)([self._mdln, self._softrev])

    def _on_s01f13(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 1, Function 13, Establish Communication Request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler, message  # unused parameters

        if self._is_host:
            return self.stream_function(1, 14)({"COMMACK": self.on_commack_requested(), "MDLN": []})

        return self.stream_function(1, 14)({"COMMACK": self.on_commack_requested(),
                                            "MDLN": [self._mdln, self._softrev]})



"""Contains hsms message class."""
from __future__ import annotations

import secsgem.common

from .header import HsmsHeader


class HsmsBlock(secsgem.common.Block[HsmsHeader]):
    """Data block for SECS I."""

    header_type = HsmsHeader
    length_format = "L"
    checksum_format = ""


class HsmsMessage(secsgem.common.Message):
    """Class for hsms message.

    Contains all required data and functions.
    """

    block_size = -1
    block_type = HsmsBlock

    @property
    def header(self) -> HsmsHeader:
        """Get the header."""
        return self._blocks[0].header

    @property
    def data(self) -> bytes:
        """Get the data."""
        return self._blocks[0].data

    @property
    def complete(self) -> bool:
        """Check if the message is complete."""
        return True



"""Class for stream 02 function 43."""

from secsgem.secs.data_items import FCNID, STRID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F43(SecsStreamFunction):
    """reset spooling streams and functions - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F43
        [
            {
                STRID: U1[1]
                FCNID: [
                    DATA: U1[1]
                    ...
                ]
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F43([{"STRID": 1, "FCNID": [10, 20]}, {"STRID": 2, "FCNID": [30, 40]}])
        S2F43 W
          <L [2]
            <L [2]
              <U1 1 >
              <L [2]
                <U1 10 >
                <U1 20 >
              >
            >
            <L [2]
              <U1 2 >
              <L [2]
                <U1 30 >
                <U1 40 >
              >
            >
          > .

    Data Items:
        - :class:`STRID <secsgem.secs.data_items.STRID>`
        - :class:`FCNID <secsgem.secs.data_items.FCNID>`

    """

    _stream = 2
    _function = 43

    _data_format = [
        [
            STRID,
            [FCNID]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Data Value capability."""
from __future__ import annotations

import typing

from .capability import Capability
from .handler import GemHandler

if typing.TYPE_CHECKING:
    import secsgem.secs

    from .data_value import DataValue


class DataValueCapability(GemHandler, Capability):
    """Data Value capability on GEM equipment."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self.__data_values: dict[int | str, DataValue] = {
        }

    @property
    def _data_values(self) -> dict[int | str, DataValue]:
        """Get list of the data values.

        Returns:
            Data value list

        """
        return self.__data_values

    @property
    def data_values(self) -> dict[int | str, DataValue]:
        """Get list of the data values.

        Returns:
            Data value list

        """
        return self._data_values

    def on_dv_value_request(self,
                            data_value_id: secsgem.secs.variables.Base,
                            data_value: DataValue) -> secsgem.secs.variables.Base:
        """Get the data value depending on its configuation.

        Override in inherited class to provide custom data value request handling.

        Args:
            data_value_id: Id of the data value encoded in the corresponding type
            data_value: The data value requested

        Returns:
            The value encoded in the corresponding type

        """
        del data_value_id  # unused variable

        return data_value.value_type(data_value.value)

    def _get_dv_value(self, data_value: DataValue) -> secsgem.secs.variables.Base:
        """Get the data value depending on its configuation.

        Args:
            data_value: The data value requested

        Returns:
            The value encoded in the corresponding type

        """
        if data_value.use_callback:
            return self.on_dv_value_request(data_value.id_type(data_value.dvid), data_value)

        return data_value.value_type(data_value.value)



"""Class for stream 10 function 02."""

from secsgem.secs.data_items import ACKC10
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS10F02(SecsStreamFunction):
    """terminal - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS10F02
        ACKC10: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS10F02(secsgem.secs.data_items.ACKC10.ACCEPTED)
        S10F2
          <B 0x0> .

    Data Items:
        - :class:`ACKC10 <secsgem.secs.data_items.ACKC10>`

    """

    _stream = 10
    _function = 2

    _data_format = ACKC10

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""State machine for connection state."""
from __future__ import annotations

import logging
import typing
from typing import Any

from .events import EventProducer

if typing.TYPE_CHECKING:
    import enum


class UnknownTransitionError(Exception):
    """Exception for unknown transition."""

    def __init__(self, transition: str) -> None:
        """Initialize unknown transition exception.

        Args:
            transition: name of the transition

        """
        super().__init__(f"Invalid transition: {transition}")


class WrongSourceStateError(Exception):
    """Exception for wrong source state for transition."""

    def __init__(self, transition: str, expected: str, actual: str) -> None:
        """Initialize wrong source state exception.

        Args:
            transition: name of the transition
            expected: expected source state
            actual: actual source state

        """
        super().__init__(f"Invalid source state for transition '{transition}': {actual} (expected {expected})")


class State:
    """State machine state class."""

    def __init__(
        self,
        state: enum.Enum,
        name: str,
        parent: State | None = None,
        initial: bool = False,
    ) -> None:
        """Initialize state object.

        Args:
            state: state
            name: state name
            parent: parent state
            initial: is initial state

        """
        self._state = state
        self._name = name
        self._parent = parent
        self._active = initial

        self._event_producer = EventProducer()

    @property
    def state(self) -> enum.Enum:
        """Get the connection state for this state."""
        return self._state

    @property
    def name(self) -> str:
        """Get the name for this state."""
        return self._name

    @property
    def events(self) -> EventProducer:
        """Property for event handling."""
        return self._event_producer

    @property
    def parent(self) -> State | None:
        """Get parent state if available."""
        return self._parent

    @property
    def active(self) -> bool:
        """Get if the state is active."""
        return self._active

    def enter(self, source: State | None):
        """Enter the state.

        Args:
            source: state to enter from.

        """
        self.events.fire("enter", {})

        self._active = True

        if self.parent is not None and (source is None or source.parent != self.parent):
            self.parent.enter(source.parent if source is not None else None)

    def leave(self, destination: State | None):
        """Leave the state.

        Args:
            destination: state to enter from.

        """
        self.events.fire("leave", {})

        self._active = False

        if self.parent is not None and (destination is None or destination.parent != self.parent):
            self.parent.leave(destination.parent if destination is not None else None)


class Transition:
    """State machine transition class."""

    def __init__(
        self,
        name: str,
        sources: State | list[State],
        destination: State
    ) -> None:
        """Initialize transition object.

        Args:
            name: name of the transition
            sources: source states allowed for transition
            destination: destination state for transition

        """
        self._name = name
        self._sources = sources if isinstance(sources, list) else [sources]
        self._destination = destination

        self._event_producer = EventProducer()

    @property
    def name(self) -> str:
        """Get transition name."""
        return self._name

    @property
    def events(self) -> EventProducer:
        """Property for event handling."""
        return self._event_producer

    @property
    def sources(self) -> list[State]:
        """Get the allowed source states for the transition."""
        return self._sources

    @property
    def destination(self) -> State:
        """Get the destination state for the transition."""
        return self._destination

    def __call__(self) -> Any:
        """Call the transition."""
        self.events.fire("called", {})


class StateMachine:
    """Base state machine."""

    def __init__(self) -> None:
        """Initialize state machine."""
        self._current_state: State
        self._transitions: list[Transition]
        self._logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)

    @property
    def current(self) -> enum.Enum:
        """Get the current state enum."""
        return self._current_state.state

    @property
    def current_state(self) -> State:
        """Get the current state."""
        return self._current_state

    def transition(self, name: str) -> Transition:
        """Get the object for a specific transition.

        Args:
            name: transition name

        """
        value = next((transition for transition in self._transitions if transition.name == name), None)

        if value is None:
            raise UnknownTransitionError(name)

        return value


    def _perform_transition(self, name: str) -> None:
        """Perform a transition.

        Args:
            name: transition name

        """
        transition = self.transition(name)

        if self._current_state not in transition.sources:
            raise WrongSourceStateError(
                name,
                "/".join([state.name for state in transition.sources]),
                self._current_state.name
            )

        self._logger.debug("State change: %s >> %s", self._current_state.name, transition.destination.name)
        self._current_state.leave(transition.destination)

        old_state = self._current_state
        self._current_state = transition.destination

        transition.destination.enter(old_state)

        transition()



"""Class for stream 06 function 16."""

from secsgem.secs.data_items import CEID, DATAID, RPTID, V
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F16(SecsStreamFunction):
    """event report data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F16
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            CEID: U1/U2/U4/U8/I1/I2/I4/I8/A
            RPT: [
                {
                    RPTID: U1/U2/U4/U8/I1/I2/I4/I8/A
                    V: [
                        DATA: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                        ...
                    ]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F16({
        ...     "DATAID": 1,
        ...     "CEID": 1337,
        ...     "RPT": [{
        ...         "RPTID": 1000,
        ...         "V": ["VAR", secsgem.secs.variables.U4(100)]}]})
        S6F16
          <L [3]
            <U1 1 >
            <U2 1337 >
            <L [1]
              <L [2]
                <U2 1000 >
                <L [2]
                  <A "VAR">
                  <U4 100 >
                >
              >
            >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`CEID <secsgem.secs.data_items.CEID>`
        - :class:`RPTID <secsgem.secs.data_items.RPTID>`
        - :class:`V <secsgem.secs.data_items.V>`

    """

    _stream = 6
    _function = 16

    _data_format = [
        DATAID,
        CEID,
        [
            [
                "RPT",
                RPTID,
                [V]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Contains callback handling routines."""
from __future__ import annotations

import typing


class _CallbackCallWrapper:
    def __init__(self, handler: CallbackHandler, name: str):
        self._name = name
        self._handler = handler

    def __call__(self, *args: tuple, **kwargs: dict) -> typing.Any:
        return self.handler._call(self.name, *args, **kwargs)  # noqa: SLF001

    @property
    def name(self) -> str:
        """Get the callback name."""
        return self._name

    @property
    def handler(self) -> CallbackHandler:
        """Get the handler for the callback."""
        return self._handler


class CallbackHandler:
    """Handler for callbacks for HSMS/SECS/GEM events.

    This handler manages callbacks for events that can happen on a handler for a connection.
    """

    def __init__(self) -> None:
        """Initialize the handler."""
        self._callbacks: dict[str, typing.Callable] = {}
        self.target: object = None
        self._object_intitialized = True

    def __setattr__(self, name: str, value: typing.Callable):
        """Set an item as object member.

        Args:
            name: name of the callback
            value: callback function

        """
        if "_object_intitialized" not in self.__dict__ or name in self.__dict__:
            dict.__setattr__(self, name, value)
            return

        if value is None:
            if name in self._callbacks:
                del self._callbacks[name]
        else:
            self._callbacks[name] = value

    def __getattr__(self, name: str) -> typing.Callable:
        """Get a callable function for an event.

        Args:
            name: name of the event

        """
        return _CallbackCallWrapper(self, name)

    class _CallbacksIter:
        def __init__(self, keys):
            self._keys = list(keys)
            self._counter = 0

        def __iter__(self):
            return self

        def __next__(self):
            if self._counter < len(self._keys):
                i = self._counter
                self._counter += 1
                return self._keys[i]

            raise StopIteration

    def __iter__(self) -> _CallbacksIter:
        """Get an iterator for the callbacks.

        Returns:
            callback iterator

        """
        return self._CallbacksIter(self._callbacks.keys())

    def __contains__(self, callback: str) -> bool:
        """Check if a callback is present.

        Args:
            callback: name of the event

        Returns:
            True if callback present

        """
        if callback in self._callbacks:
            return True

        delegate_handler = getattr(self.target, "_on_" + callback, None)
        if callable(delegate_handler):
            return True

        return False

    def _call(self, callback: str, *args, **kwargs) -> typing.Any:
        if callback in self._callbacks:
            return self._callbacks[callback](*args, **kwargs)

        delegate_handler = getattr(self.target, "_on_" + callback, None)
        if callable(delegate_handler):
            return delegate_handler(*args, **kwargs)

        return None



"""module imports."""

from secsgem.common.settings import DeviceType

from .deselect_req_header import HsmsDeselectReqHeader
from .deselect_rsp_header import HsmsDeselectRspHeader
from .header import HsmsHeader, HsmsSType
from .linktest_req_header import HsmsLinktestReqHeader
from .linktest_rsp_header import HsmsLinktestRspHeader
from .message import HsmsBlock, HsmsMessage
from .protocol import HsmsProtocol
from .reject_req_header import HsmsRejectReqHeader
from .select_req_header import HsmsSelectReqHeader
from .select_rsp_header import HsmsSelectRspHeader
from .separate_req_header import HsmsSeparateReqHeader
from .settings import HsmsConnectMode, HsmsSettings
from .stream_function_header import HsmsStreamFunctionHeader

__all__ = [
    "HsmsProtocol",
    "HsmsMessage", "HsmsBlock",
    "HsmsStreamFunctionHeader", "HsmsSeparateReqHeader", "HsmsRejectReqHeader", "HsmsLinktestRspHeader",
    "HsmsLinktestReqHeader", "HsmsDeselectRspHeader", "HsmsDeselectReqHeader", "HsmsSelectRspHeader",
    "HsmsSelectReqHeader", "HsmsHeader", "HsmsSType",
    "HsmsSettings", "HsmsConnectMode", "DeviceType",
]



"""Class for stream 09 function 09."""

from secsgem.secs.data_items import SHEAD
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS09F09(SecsStreamFunction):
    """transaction timer timeout.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS09F09
        SHEAD: B[10]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS09F09("HEADERDATA")
        S9F9
          <B 0x48 0x45 0x41 0x44 0x45 0x52 0x44 0x41 0x54 0x41> .

    Data Items:
        - :class:`SHEAD <secsgem.secs.data_items.SHEAD>`

    """

    _stream = 9
    _function = 9

    _data_format = SHEAD

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Wrapper for GEM equipment constant."""
from __future__ import annotations

import enum

import secsgem.secs


class EquipmentConstantId(enum.Enum):
    """Default IDs for equipment constants."""
    ESTABLISH_COMMUNICATIONS_TIMEOUT = 1
    TIME_FORMAT = 2


class EquipmentConstant:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """Equipment constant definition."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        ecid: int | str | EquipmentConstantId,
        name: str,
        min_value: int | float,
        max_value: int | float,
        default_value: int | float,
        unit: str,
        value_type: type[secsgem.secs.variables.Base],
        use_callback: bool = True,
        **kwargs,
    ):
        """Initialize an equipment constant.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Custom parameters can be set with the keyword arguments,
        they will be passed to the GemEquipmentHandlers callbacks
        :func:`secsgem.gem.equipmenthandler.GemEquipmentHandler.on_ec_value_request`
        and :func:`secsgem.gem.equipmenthandler.GemEquipmentHandler.on_ec_value_update` .

        If use_callbacks is disabled, you can set the value with the value property.

        Args:
            ecid: ID of the equipment constant
            name: long name
            min_value: minimum value
            max_value: maximum value
            default_value: default value
            unit: unit (see SEMI E5, Units of Measure)
            value_type: type of the status variable
            use_callback: use the GemEquipmentHandler callbacks to get and set variable (True) or use internal value
            **kwargs: additional attributes for object

        """
        self.ecid = ecid if not isinstance(ecid, EquipmentConstantId) else ecid.value
        self.name = name
        self.min_value = min_value
        self.max_value = max_value
        self.default_value = default_value
        self.unit = unit
        self.value_type = value_type
        self.use_callback = use_callback
        self.value = default_value

        self.id_type: type[secsgem.secs.variables.Base]

        if isinstance(self.ecid, int):
            self.id_type = secsgem.secs.variables.U4
        else:
            self.id_type = secsgem.secs.variables.String

        for key, value in kwargs.items():
            setattr(self, key, value)



"""Class for stream 02 function 21."""

from secsgem.secs.data_items import RCMD
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F21(SecsStreamFunction):
    """Remote command send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F21
        RCMD: U1/I1/A

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F21("COMMMAND1")
        S2F21
          <A "COMMMAND1"> .

    Data Items:
        - :class:`RCMD <secsgem.secs.data_items.RCMD>`

    """

    _stream = 2
    _function = 21

    _data_format = RCMD

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""MID data item."""
from .. import variables
from .base import DataItemBase


class MID(DataItemBase):
    """Material ID.

    :Types:
       - :class:`String <secsgem.secs.variables.String>`
       - :class:`Binary <secsgem.secs.variables.Binary>`
    :Length: 80

    **Used In Function**
        - :class:`SecsS12F01 <secsgem.secs.functions.SecsS12F01>`
        - :class:`SecsS12F03 <secsgem.secs.functions.SecsS12F03>`
        - :class:`SecsS12F04 <secsgem.secs.functions.SecsS12F04>`
        - :class:`SecsS12F05 <secsgem.secs.functions.SecsS12F05>`
        - :class:`SecsS12F07 <secsgem.secs.functions.SecsS12F07>`
        - :class:`SecsS12F09 <secsgem.secs.functions.SecsS12F09>`
        - :class:`SecsS12F11 <secsgem.secs.functions.SecsS12F11>`
        - :class:`SecsS12F13 <secsgem.secs.functions.SecsS12F13>`
        - :class:`SecsS12F14 <secsgem.secs.functions.SecsS12F14>`
        - :class:`SecsS12F15 <secsgem.secs.functions.SecsS12F15>`
        - :class:`SecsS12F16 <secsgem.secs.functions.SecsS12F16>`
        - :class:`SecsS12F17 <secsgem.secs.functions.SecsS12F17>`
        - :class:`SecsS12F18 <secsgem.secs.functions.SecsS12F18>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.String,
        variables.Binary
    ]
    __count__ = 80



"""Class for stream 14 function 01."""

from secsgem.secs.data_items import ATTRDATA, ATTRID, ATTRRELN, OBJID, OBJSPEC, OBJTYPE
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS14F01(SecsStreamFunction):
    """GetAttr request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS14F01
        {
            OBJSPEC: A
            OBJTYPE: U1/U2/U4/U8/A
            OBJID: [
                DATA: U1/U2/U4/U8/A
                ...
            ]
            FILTER: [
                {
                    ATTRID: U1/U2/U4/U8/A
                    ATTRDATA: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                    ATTRRELN: U1
                }
                ...
            ]
            ATTRID: [
                DATA: U1/U2/U4/U8/A
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS14F01({
        ...     "OBJSPEC": '',
        ...     "OBJTYPE": 'StripMap',
        ...     "OBJID": ['MAP001'],
        ...     "FILTER": [],
        ...     "ATTRID": ['OriginLocation', 'Rows', 'Columns', 'CellStatus', 'LotID']})
        S14F1 W
          <L [5]
            <A>
            <A "StripMap">
            <L [1]
              <A "MAP001">
            >
            <L>
            <L [5]
              <A "OriginLocation">
              <A "Rows">
              <A "Columns">
              <A "CellStatus">
              <A "LotID">
            >
          > .

    Data Items:
        - :class:`OBJSPEC <secsgem.secs.data_items.OBJSPEC>`
        - :class:`OBJTYPE <secsgem.secs.data_items.OBJTYPE>`
        - :class:`OBJID <secsgem.secs.data_items.OBJID>`
        - :class:`ATTRID <secsgem.secs.data_items.ATTRID>`
        - :class:`ATTRDATA <secsgem.secs.data_items.ATTRDATA>`
        - :class:`ATTRRELN <secsgem.secs.data_items.ATTRRELN>`

    """

    _stream = 14
    _function = 1

    _data_format = [
        OBJSPEC,
        OBJTYPE,
        [OBJID],
        [
            [
                "FILTER",
                ATTRID,
                ATTRDATA,
                ATTRRELN
            ]
        ],
        [ATTRID]
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 02 function 34."""

from secsgem.secs.data_items import DRACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F34(SecsStreamFunction):
    """define report - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F34
        DRACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F34(secsgem.secs.data_items.DRACK.INVALID_FORMAT)
        S2F34
          <B 0x2> .

    Data Items:
        - :class:`DRACK <secsgem.secs.data_items.DRACK>`

    """

    _stream = 2
    _function = 34

    _data_format = DRACK

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 06 function 07."""

from secsgem.secs.data_items import DATAID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F07(SecsStreamFunction):
    """data transfer request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F07
        DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F07(1)
        S6F7 W
          <U1 1 > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`

    """

    _stream = 6
    _function = 7

    _data_format = DATAID

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 05 function 15."""

from secsgem.secs.data_items import ACKA, ERRCODE, ERRTEXT, EXID, TIMESTAMP
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F15(SecsStreamFunction):
    """exception recover complete - notify.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F15
        {
            TIMESTAMP: A[32]
            EXID: A[20]
            DATA: {
                ACKA: BOOLEAN[1]
                DATA: {
                    ERRCODE: I1/I2/I4/I8
                    ERRTEXT: A[120]
                }
            }
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F15({
        ...     "TIMESTAMP": "161006221500",
        ...     "EXID": "EX123",
        ...     "DATA": {
        ...         "ACKA": False,
        ...         "DATA": {
        ...             "ERRCODE": 10,
        ...             "ERRTEXT": "Error"
        ...         }}})
        S5F15
          <L [3]
            <A "161006221500">
            <A "EX123">
            <L [2]
              <BOOLEAN False >
              <L [2]
                <I1 10 >
                <A "Error">
              >
            >
          > .

    Data Items:
        - :class:`TIMESTAMP <secsgem.secs.data_items.TIMESTAMP>`
        - :class:`EXID <secsgem.secs.data_items.EXID>`
        - :class:`ACKA <secsgem.secs.data_items.ACKA>`
        - :class:`ERRCODE <secsgem.secs.data_items.ERRCODE>`
        - :class:`ERRTEXT <secsgem.secs.data_items.ERRTEXT>`

    """

    _stream = 5
    _function = 15

    _data_format = [
        TIMESTAMP,
        EXID,
        [
            ACKA,
            [
                ERRCODE,
                ERRTEXT
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""Contains helper functions."""
from __future__ import annotations

import typing


class Event:
    """Class to handle the callbacks for a single event."""

    def __init__(self) -> None:
        """Initialize the event class."""
        self._callbacks: list[typing.Callable[[dict[str, typing.Any]], None]] = []

    def register(self, callback: typing.Callable[[dict[str, typing.Any]], None]) -> None:
        """Add a new callback to event.

        Args:
            callback: function to register as callback

        """
        self._callbacks.append(callback)

    def __iadd__(self, other: typing.Callable[[dict[str, typing.Any]], None]) -> Event:
        """Add a new callback to event.

        Args:
            other: function to register as callback

        Returns:
            updated instance

        """
        self.register(other)
        return self

    def unregister(self, callback: typing.Callable[[dict[str, typing.Any]], None]) -> None:
        """Remove a callback from the event.

        Args:
            callback: function to unregister

        """
        self._callbacks.remove(callback)

    def __isub__(self, other: typing.Callable[[dict[str, typing.Any]], None]) -> Event:
        """Remove a callback from event.

        Args:
            other: function to unregister

        Returns:
            updated instance

        """
        self.unregister(other)
        return self

    def __call__(self, data: dict[str, typing.Any]):
        """Raise the event and call all callbacks."""
        for callback in self._callbacks:
            callback(data)

    def __len__(self) -> int:
        """Return the number of callbacks."""
        return len(self._callbacks)

    def __repr__(self) -> str:
        """Generate representation for an object."""
        return f"{self.__class__.__name__}: {self._callbacks}"


class Targets:
    """Class to handle a list of objects as target for events."""

    def __init__(self) -> None:
        """Initialize the target class."""
        self._targets: list[object] = []

    def __iadd__(self, other: object) -> Targets:
        """Add a targets."""
        self._targets.append(other)
        return self

    def __isub__(self, other: object) -> Targets:
        """Remove a target."""
        self._targets.remove(other)
        return self

    class _TargetsIter:
        def __init__(self, values):
            self._values = values
            self._counter = 0

        def __iter__(self):
            """Return the iterator."""
            return self

        def __next__(self):
            """Get the next item or raise StopIteration if at end of list."""
            if self._counter < len(self._values):
                i = self._counter
                self._counter += 1
                return self._values[i]

            raise StopIteration

    def __iter__(self) -> _TargetsIter:
        """Return the iterator."""
        return self._TargetsIter(self._targets)


class EventProducer:
    """Manages the consumers for the events and handles firing events."""

    def __init__(self) -> None:
        """Initialize the event producer class."""
        self._targets = Targets()
        self._events: dict[str, Event] = {}

    def __getattr__(self, name: str) -> Event:
        """Get an event as member of the EventProducer object."""
        if name not in self._events:
            self._events[name] = Event()

        return self._events[name]

    def __iadd__(self, other) -> EventProducer:
        """Add a the callbacks and targets of another EventProducer to this one."""
        for event_name in other._events:
            if event_name not in self._events:
                self._events[event_name] = Event()

            for callback in other._events[event_name]._callbacks:
                self._events[event_name] += callback

        for target in other._targets:
            self._targets += target
        return self

    def fire(self, event: str, data: dict[str, typing.Any]):
        """Fire a event.

        calls all the available handlers for a specific event

        Args:
            event: name of the event
            data: data connected to this event

        """
        for target in self._targets:
            generic_handler = getattr(target, "_on_event", None)
            if callable(generic_handler):
                generic_handler(event, data)

            specific_handler = getattr(target, "_on_event_" + event, None)
            if callable(specific_handler):
                specific_handler(data)

        if event in self._events:
            self._events[event](data)

    def __repr__(self) -> str:
        """Generate representation for an object."""
        return f"{self.__class__.__name__}: {self._events}"

    class _EventsIter:
        def __init__(self, keys):
            self._keys = list(keys)
            self._counter = 0

        def __iter__(self):
            """Return the iterator."""
            return self

        def __next__(self):
            """Get the next item or raise StopIteration if at end of list."""
            if self._counter < len(self._keys):
                i = self._counter
                self._counter += 1
                return self._keys[i]

            raise StopIteration

    def __iter__(self) -> _EventsIter:
        """Return the iterator."""
        return self._EventsIter([event for event, event_value in self._events.items() if len(event_value) > 0])

    @property
    def targets(self) -> Targets:
        """Targets used as consumer for this producer."""
        return self._targets

    @targets.setter
    def targets(self, value: Targets):
        if self._targets != value:
            raise AttributeError("can't set attribute")



"""SECS 4 byte unsigned integer variable type."""

from .base_number import BaseNumber


class U4(BaseNumber):
    """Secs type for 4 byte unsigned data.

    :param value: initial value
    :type value: list/integer
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o54
    text_code = "U4"
    _base_type = int
    _min = 0
    _max = 4294967295
    _bytes = 4
    _struct_code = "L"
    preferred_types = [int]



"""Base class for for SECS stream and functions."""
from __future__ import annotations

import typing

import secsgem.common

from ..data_items import DataItemBase
from ..variables import functions

DataItemRecursive = typing.Union[typing.Type[DataItemBase], typing.Iterable["DataItemRecursive"]]


class StructureDisplayingMeta(type):
    """Meta class overriding the default __repr__ of a class."""

    def __repr__(cls):
        """Generate textual representation for an object of this class."""
        return cls.get_format()


class _ClassProperty(property):
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class SecsStreamFunction(metaclass=StructureDisplayingMeta):  # pylint: disable=too-many-instance-attributes
    """Secs stream and function base class.

    This class is inherited to create a stream/function class.
    To create a function specific content the class variables :attr:`_stream`, :attr:`_function`
    and :attr:`_data_format` must be overridden.
    """

    _stream = 0
    _function = 0

    _data_format: DataItemRecursive | None = None

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False

    def __init__(self, value=None):
        """Initialize a stream function object.

        Example:
            class SecsS02F30(SecsStreamFunction):
                _stream = 2
                _function = 30

                _to_host = True
                _to_equipment = False

                _has_reply = False
                _is_reply_required = False

                _is_multi_block = True

                _data_format = [
                    [
                        ECID,
                        ECNAME,
                        ECMIN,
                        ECMAX,
                        ECDEF,
                        UNITS
                    ]
                ]

        :param value: set the value of stream/function parameters
        :type value: various
        """
        self.data = functions.generate(self._data_format)

        self.data_format = self._data_format
        self.to_host = self._to_host
        self.to_equipment = self._to_equipment

        self.has_reply = self._has_reply
        self.is_reply_required = self._is_reply_required

        self.is_multi_block = self._is_multi_block

        if value is not None and self.data is not None:
            self.data.set(value)

        self._object_intitialized = True

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        function = f"S{self.stream}F{self.function}"
        if self.data is None:
            return f"{function}{' W' if self._is_reply_required else ''} ."
        data = f"{self.data}"

        return f"{function}{' W' if self._is_reply_required else ''}\n{secsgem.common.indent_block(data)} ."

    def __getitem__(self, key):
        """Get an item using the indexer operator."""
        return self.data[key]

    def __setitem__(self, key, item):
        """Set an item using the indexer operator."""
        self.data[key] = item

    def __len__(self):
        """Get the length."""
        return len(self.data)

    def __getattr__(self, item):
        """Get an item as object member."""
        return self.data.__getattr__(item)

    def __setattr__(self, item, value):
        """Set an item as object member."""
        if "_object_intitialized" not in self.__dict__:
            dict.__setattr__(self, item, value)
            return

        if item in self.data.data:
            self.data.__setattr__(item, value)
            return

    def append(self, data):
        """Append data to list, if stream/function parameter is a list.

        :param data: list item to add
        :type data: various
        """
        if hasattr(self.data, "append") and callable(self.data.append):
            self.data.append(data)
        else:
            raise AttributeError(
                f"class {self.__class__.__name__} has no attribute 'append'")

    def encode(self):
        """Generate the encoded hsms data of the stream/function parameter.

        :returns: encoded data
        :rtype: string
        """
        if self.data is None:
            return b""

        return self.data.encode()

    def decode(self, data):
        """Update stream/function parameter data from the passed data.

        :param data: encoded data
        :type data: string
        """
        if self.data is not None:
            self.data.decode(data)

    def set(self, value):
        """Update the value of the stream/function parameter.

        :param value: new value for the parameter
        :type value: various
        """
        self.data.set(value)

    def get(self):
        """Get the current value of the stream/function parameter.

        :returns: current parameter value
        :rtype: various
        """
        if self.data is None:
            return None

        return self.data.get()

    @classmethod
    def get_format(cls):
        """Get the format of the function.

        :returns: returns the string representation of the function
        :rtype: string
        """
        if cls._data_format is not None:
            return functions.get_format(cls._data_format)

        return "Header only"

    @_ClassProperty
    def stream(self) -> int:
        """Get the stream number of this function."""
        return self._stream

    @_ClassProperty
    def function(self) -> int:
        """Get the function number of this function."""
        return self._function



"""settings base class."""
from __future__ import annotations

import abc
import enum
import typing

from .timeouts import Timeouts

if typing.TYPE_CHECKING:
    from secsgem.secs.functions import StreamsFunctions

    from .connection import Connection
    from .protocol import Protocol


class DeviceType(enum.Enum):
    """Type of device (equipment or host)."""

    # This device is equipment
    EQUIPMENT = 0

    # This device is host.
    HOST = 1


class Settings(abc.ABC):
    r"""Settings base class.

    These attributes can be initialized in the constructor and accessed as property.

    """

    def __init__(self, **kwargs) -> None:
        """Initialize settings."""
        from secsgem.secs.functions import StreamsFunctions  # pylint: disable=import-outside-toplevel

        self._timeouts = Timeouts(**kwargs)
        self._device_type = kwargs.get("device_type", DeviceType.HOST)
        self._streams_functions = kwargs.get("streams_functions", StreamsFunctions())
        self._session_id = kwargs.get("session_id", 0)
        self._establish_communication_timeout = kwargs.get("establish_communication_timeout", 10)

    @property
    def address(self):
        raise NotImplementedError("针对hsms需要继承实现这个函数.")

    @property
    def port(self):
        raise NotImplementedError("针对hsms需要继承实现这个函数.")

    @property
    def timeouts(self) -> Timeouts:
        """Communication timeouts."""
        return self._timeouts

    @property
    def device_type(self) -> DeviceType:
        """Device type.

        Default: DeviceType.HOST.
        """
        return self._device_type

    @property
    def streams_functions(self) -> StreamsFunctions:
        """Container with streams/functions.

        Default: DeviceType.HOST.
        """
        return self._streams_functions

    @property
    def session_id(self) -> int:
        """Session / device ID to use for connection.

        Default: 0
        """
        return self._session_id

    @property
    def establish_communication_timeout(self) -> int:
        """Time to wait between CA requests.

        Default: 10
        """
        return self._establish_communication_timeout

    @establish_communication_timeout.setter
    def establish_communication_timeout(self, value: int) -> None:
        """Set time to wait between CA requests."""
        self._establish_communication_timeout = value

    @abc.abstractmethod
    def create_protocol(self) -> Protocol:
        """Protocol class for this configuration."""
        raise NotImplementedError(f"function 'create_protocol' is not implemented for '{self.__class__.__name__}'")

    @abc.abstractmethod
    def create_connection(self) -> Connection:
        """Connection class for this configuration."""
        raise NotImplementedError(f"function 'create_connection' is not implemented for '{self.__class__.__name__}'")

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """Name of this configuration."""
        raise NotImplementedError(f"property 'name' is not implemented for '{self.__class__.__name__}'")

    @abc.abstractmethod
    def generate_thread_name(self, functionality: str) -> str:
        """Generate a unique thread name for this configuration and a provided functionality.

        Args:
            functionality: name of the functionality to generate thread name for

        Returns:
            generated thread name

        """
        raise NotImplementedError(f"function 'generate_thread_name' is not implemented for '{self.__class__.__name__}'")



"""Equipment Constants capability."""
from __future__ import annotations

import secsgem.secs

from .capability import Capability
from .equipment_constant import EquipmentConstant, EquipmentConstantId
from .handler import GemHandler


class EquipmentConstantsCapability(GemHandler, Capability):
    """Equipment Contstants capability on GEM equipment."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self._equipment_constants: dict[int | str | EquipmentConstantId, EquipmentConstant] = {
            EquipmentConstantId.ESTABLISH_COMMUNICATIONS_TIMEOUT.value: EquipmentConstant(
                EquipmentConstantId.ESTABLISH_COMMUNICATIONS_TIMEOUT,
                "EstablishCommunicationsTimeout",
                10,
                120,
                10,
                "sec",
                secsgem.secs.variables.I2),
            EquipmentConstantId.TIME_FORMAT.value: EquipmentConstant(
                EquipmentConstantId.TIME_FORMAT,
                "TimeFormat",
                0,
                2,
                1,
                "",
                secsgem.secs.variables.I4),
        }

    @property
    def equipment_constants(self) -> dict[int | str | EquipmentConstantId, EquipmentConstant]:
        """The list of the equipments contstants.

        Returns:
            Equipment constant list

        """
        return self._equipment_constants

    def on_ec_value_request(self,
                            equipment_constant_id: secsgem.secs.variables.Base,
                            equipment_constant: EquipmentConstant) -> secsgem.secs.variables.Base:
        """Get the equipment constant value depending on its configuation.

        Override in inherited class to provide custom equipment constant request handling.

        Args:
            equipment_constant_id: Id of the equipment constant encoded in the corresponding type
            equipment_constant: The equipment constant requested

        Returns:
            The value encoded in the corresponding type

        """
        del equipment_constant_id  # unused variable

        return equipment_constant.value_type(equipment_constant.value)

    def on_ec_value_update(self,
                           equipment_constant_id: secsgem.secs.variables.Base,
                           equipment_constant: EquipmentConstant,
                           value: int | float):
        """Set the equipment constant value depending on its configuation.

        Override in inherited class to provide custom equipment constant update handling.

        Args:
            equipment_constant_id: Id of the equipment constant encoded in the corresponding type
            equipment_constant: The equipment constant to be updated
            value: The value encoded in the corresponding type

        """
        del equipment_constant_id  # unused variable

        equipment_constant.value = value

    def _get_ec_value(self, equipment_constant: EquipmentConstant) -> secsgem.secs.variables.Base:
        """Get the equipment constant value depending on its configuation.

        Args:
            equipment_constant: The equipment requested

        Returns:
            The value encoded in the corresponding type

        """
        if equipment_constant.ecid == EquipmentConstantId.ESTABLISH_COMMUNICATIONS_TIMEOUT.value:
            return equipment_constant.value_type(self.settings.establish_communication_timeout)
        if equipment_constant.ecid == EquipmentConstantId.TIME_FORMAT.value:
            return equipment_constant.value_type(self._time_format)

        if equipment_constant.use_callback:
            return self.on_ec_value_request(equipment_constant.id_type(equipment_constant.ecid), equipment_constant)
        return equipment_constant.value_type(equipment_constant.value)

    def _set_ec_value(self, equipment_constant: EquipmentConstant, value: int | float):
        """Get the equipment constant value depending on its configuation.

        Args:
            equipment_constant: The equipment requested
            value: The value encoded in the corresponding type

        """
        if equipment_constant.ecid == EquipmentConstantId.ESTABLISH_COMMUNICATIONS_TIMEOUT.value:
            self.settings.establish_communication_timeout = int(value)
        if equipment_constant.ecid == EquipmentConstantId.TIME_FORMAT.value:
            self._time_format = int(value)

        if equipment_constant.use_callback:
            self.on_ec_value_update(equipment_constant.id_type(equipment_constant.ecid), equipment_constant, value)
        else:
            equipment_constant.value = value

    def _on_s02f13(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 2, Function 13, Equipment constant request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        responses = []

        if len(function) == 0:
            responses = [self._get_ec_value(equipment_constant)
                         for equipment_constant in self._equipment_constants.values()]
        else:
            for equipment_constant_id in function:  # type: ignore[attr-defined]
                if equipment_constant_id not in self._equipment_constants:
                    responses.append(secsgem.secs.variables.Array(secsgem.secs.data_items.ECV, []))
                else:
                    equipment_constant = self._equipment_constants[equipment_constant_id]
                    responses.append(self._get_ec_value(equipment_constant))

        return self.stream_function(2, 14)(responses)

    def _on_s02f15(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 2, Function 15, Equipment constant send.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        eac = 0

        for equipment_constant in function:  # type: ignore[attr-defined]
            if equipment_constant.ECID not in self._equipment_constants:
                eac = 1
            else:
                constant = self.equipment_constants[equipment_constant.ECID.get()]

                if constant.min_value is not None and equipment_constant.ECV.get() < constant.min_value:
                    eac = 3

                if constant.max_value is not None and equipment_constant.ECV.get() > constant.max_value:
                    eac = 3

        if eac == 0:
            for equipment_constant in function:  # type: ignore[attr-defined]
                self._set_ec_value(self._equipment_constants[equipment_constant.ECID], equipment_constant.ECV.get())

        return self.stream_function(2, 16)(eac)

    def _on_s02f29(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 2, Function 29, EC namelist request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        responses = []

        if len(function) == 0:
            responses = [{
                "ECID": eq_constant.ecid,
                "ECNAME": eq_constant.name,
                "ECMIN": eq_constant.min_value if eq_constant.min_value is not None else "",
                "ECMAX": eq_constant.max_value if eq_constant.max_value is not None else "",
                "ECDEF": eq_constant.default_value,
                "UNITS": eq_constant.unit,
            } for eq_constant in self._equipment_constants.values()]
        else:
            for ecid in function:  # type: ignore[attr-defined]
                if ecid not in self._equipment_constants:
                    responses.append({"ECID": ecid, "ECNAME": "", "ECMIN": "", "ECMAX": "", "ECDEF": "", "UNITS": ""})
                else:
                    eq_constant = self._equipment_constants[ecid]
                    responses.append({"ECID": eq_constant.ecid, "ECNAME": eq_constant.name,
                                      "ECMIN": eq_constant.min_value if eq_constant.min_value is not None else "",
                                      "ECMAX": eq_constant.max_value if eq_constant.max_value is not None else "",
                                      "ECDEF": eq_constant.default_value, "UNITS": eq_constant.unit})

        return self.stream_function(2, 30)(responses)



"""Class for stream 05 function 09."""

from secsgem.secs.data_items import EXID, EXMESSAGE, EXRECVRA, EXTYPE, TIMESTAMP
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F09(SecsStreamFunction):
    """exception post - notify.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F09
        {
            TIMESTAMP: A[32]
            EXID: A[20]
            EXTYPE: A
            EXMESSAGE: A
            EXRECVRA: [
                DATA: A[40]
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F09({
        ...     "TIMESTAMP": "161006221500",
        ...     "EXID": "EX123",
        ...     "EXTYPE": "ALARM",
        ...     "EXMESSAGE": "Exception",
        ...     "EXRECVRA": ["EXRECVRA1", "EXRECVRA2"] })
        S5F9
          <L [5]
            <A "161006221500">
            <A "EX123">
            <A "ALARM">
            <A "Exception">
            <L [2]
              <A "EXRECVRA1">
              <A "EXRECVRA2">
            >
          > .

    Data Items:
        - :class:`TIMESTAMP <secsgem.secs.data_items.TIMESTAMP>`
        - :class:`EXID <secsgem.secs.data_items.EXID>`
        - :class:`EXTYPE <secsgem.secs.data_items.EXTYPE>`
        - :class:`EXMESSAGE <secsgem.secs.data_items.EXMESSAGE>`
        - :class:`EXRECVRA <secsgem.secs.data_items.EXRECVRA>`

    """

    _stream = 5
    _function = 9

    _data_format = [
        TIMESTAMP,
        EXID,
        EXTYPE,
        EXMESSAGE,
        [EXRECVRA]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""SECS list variable type."""

from collections import OrderedDict

import secsgem.common

from .base import Base


class List(Base):
    """List variable type. List with items of different types."""

    format_code = 0
    text_code = "L"
    preferred_types = [dict]

    class _SecsVarListIter:
        def __init__(self, keys):
            self._keys = list(keys)
            self._counter = 0

        def __iter__(self):
            """Get an iterator."""
            return self

        def __next__(self):
            """Get the next item or raise StopIteration if at end of list."""
            if self._counter < len(self._keys):
                i = self._counter
                self._counter += 1
                return self._keys[i]

            raise StopIteration

    def __init__(self, data_format, value=None):
        """Initialize a secs list variable.

        :param data_format: internal data values
        :type data_format: OrderedDict
        :param value: initial value
        :type value: dict/list
        :param count: number of fields in the list
        :type count: integer
        """
        super().__init__()

        self.name = "DATA"

        self.data = self._generate(data_format)

        if value is not None:
            self.set(value)

        self._object_intitialized = True

    @staticmethod
    def get_format(data_format, showname=False):
        """Get the format of the variable.

        :returns: returns the string representation of the function
        :rtype: string
        """
        from .array import Array  # pylint: disable=import-outside-toplevel,cyclic-import

        array_name = f"{List.get_name_from_format(data_format)}: " if showname else ""

        if isinstance(data_format, list):
            items = []
            for item in data_format:
                if isinstance(item, str):
                    continue
                if isinstance(item, list):
                    if len(item) == 1:
                        items.append(secsgem.common.indent_block(Array.get_format(item[0], True), 4))
                    else:
                        items.append(secsgem.common.indent_block(List.get_format(item, True), 4))
                else:
                    items.append(secsgem.common.indent_block(item.get_format(), 4))
            return array_name + "{\n" + "\n".join(items) + "\n}"
        return None

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        if len(self.data) == 0:
            return f"<{self.text_code}>"

        data = ""

        for field_name in self.data:
            data += f"{secsgem.common.indent_block(self.data[field_name].__repr__())}\n"

        return f"<{self.text_code} [{len(self.data)}]\n{data}\n>"

    def __len__(self):
        """Get the length."""
        return len(self.data)

    def __getitem__(self, index):
        """Get an item using the indexer operator."""
        if isinstance(index, int):
            return self.data[list(self.data.keys())[index]]
        return self.data[index]

    def __iter__(self):
        """Get an iterator."""
        return List._SecsVarListIter(self.data.keys())

    def __setitem__(self, index, value):
        """Set an item using the indexer operator."""
        if isinstance(index, int):
            index = list(self.data.keys())[index]

        if isinstance(value, (type(self.data[index]), self.data[index].__class__.__bases__)):
            self.data[index] = value
        elif isinstance(value, Base):
            raise TypeError(
                f"Wrong type {value.__class__.__name__} when expecting {self.data[index].__class__.__name__}")
        else:
            self.data[index].set(value)

    def _generate(self, data_format):
        from .array import Array  # pylint: disable=import-outside-toplevel,cyclic-import
        from .functions import generate  # pylint: disable=import-outside-toplevel,cyclic-import

        if data_format is None:
            return None

        result_data = OrderedDict()
        for item in data_format:
            if isinstance(item, str):
                self.name = item
                continue

            item_value = generate(item)
            if isinstance(item_value, Array):
                result_data[item_value.name] = item_value
            elif isinstance(item_value, List):
                result_data[List.get_name_from_format(item)] = item_value
            elif isinstance(item_value, Base):
                result_data[item_value.name] = item_value
            else:
                raise TypeError(f"Can't handle item of class {data_format.__class__.__name__}")

        return result_data

    def __getattr__(self, item):
        """Get an item as member of the object."""
        try:
            return self.data.__getitem__(item)
        except KeyError as exc:
            raise AttributeError(item) from exc

    def __setattr__(self, item, value):
        """Set an item as member of the object."""
        if "_object_intitialized" not in self.__dict__:
            dict.__setattr__(self, item, value)
            return

        if item in self.data:
            if isinstance(value, (type(self.data[item]), self.data[item].__class__.__bases__)):
                self.data[item] = value
            elif isinstance(value, Base):
                raise TypeError(
                    f"Wrong type {value.__class__.__name__} when expecting {self.data[item].__class__.__name__}")
            else:
                self.data[item].set(value)
        else:
            self.__dict__.__setattr__(item, value)

    @staticmethod
    def get_name_from_format(data_format):
        """Generate a name for the passed data_format.

        :param data_format: data_format to get name for
        :type data_format: list/Base based class
        :returns: name for data_format
        :rtype: str
        """
        if not isinstance(data_format, list):
            raise TypeError(f"Can't generate item name of class {data_format.__class__.__name__}")

        if isinstance(data_format[0], str):
            return data_format[0]

        return "DATA"

    def set(self, value):
        """Set the internal value to the provided value.

        :param value: new value
        :type value: dict/list
        """
        if isinstance(value, dict):
            for field_name in value:
                self.data[field_name].set(value[field_name])
        elif isinstance(value, list):
            if len(value) > len(self.data):
                raise ValueError(f"Value has invalid field count (expected: {len(self.data)}, actual: {len(value)})")

            counter = 0
            for itemvalue in value:
                self.data[list(self.data.keys())[counter]].set(itemvalue)
                counter += 1
        else:
            raise TypeError(f"Invalid value type {type(value).__name__} for {self.__class__.__name__}")

    def get(self):
        """Return the internal value.

        :returns: internal value
        :rtype: list
        """
        data = {}
        for field_name in self.data:
            data[field_name] = self.data[field_name].get()

        return data

    def encode(self):
        """Encode the value to secs data.

        :returns: encoded data bytes
        :rtype: string
        """
        result = self.encode_item_header(len(self.data))

        for field_name in self.data:
            result += self.data[field_name].encode()

        return result

    def decode(self, data, start=0):
        """Decode the secs byte data to the value.

        :param data: encoded data bytes
        :type data: string
        :param start: start position of value the data
        :type start: integer
        :returns: new start position
        :rtype: integer
        """
        (text_pos, _, length) = self.decode_item_header(data, start)

        # list
        for i in range(length):
            field_name = list(self.data.keys())[i]
            text_pos = self.data[field_name].decode(data, text_pos)

        return text_pos



"""SECS variable types."""

from .array import Array
from .base import Base
from .binary import Binary
from .boolean import Boolean
from .dynamic import Dynamic
from .f4 import F4
from .f8 import F8
from .i1 import I1
from .i2 import I2
from .i4 import I4
from .i8 import I8
from .jis8 import JIS8
from .list_type import List
from .string import String
from .u1 import U1
from .u2 import U2
from .u4 import U4
from .u8 import U8

__all__ = [
    "Base",
    "Dynamic",
    "Array",
    "List",
    "Binary",
    "Boolean",
    "String",
    "JIS8",
    "F4",
    "F8",
    "I1",
    "I2",
    "I4",
    "I8",
    "U1",
    "U2",
    "U4",
    "U8",
]



from secsgem.secs.functions.base import SecsStreamFunction
from secsgem.secs.data_items import TIME


class SecsS02F31(SecsStreamFunction):
    _stream = 2
    _function = 31

    _data_format = TIME

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""TIME data item."""
from .. import variables
from .base import DataItemBase


class TIME(DataItemBase):
    """Time of day.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 32

    **Used In Function**
        - :class:`SecsS02F18 <secsgem.secs.functions.SecsS02F18>`

    """

    __type__ = variables.String
    __count__ = 32



"""module imports."""

from . import data_items, functions, variables
from .functions.base import SecsStreamFunction
from .handler import SecsHandler

__all__ = ["variables", "data_items", "functions", "SecsStreamFunction", "SecsHandler"]



"""Wrapper for GEM collection event."""
from __future__ import annotations

import enum

import secsgem.secs.variables


class CollectionEventId(enum.Enum):
    """Default IDs for collection events."""
    EQUIPMENT_OFFLINE = 1
    CONTROL_STATE_LOCAL = 2
    CONTROL_STATE_REMOTE = 3
    CMD_START_DONE = 20
    CMD_STOP_DONE = 21


class CollectionEvent:  # pylint: disable=too-few-public-methods
    """Collection event definition."""

    def __init__(self,
                 ceid: int | str | CollectionEventId,
                 name: str,
                 data_values: list[int | str],
                 **kwargs):
        """Initialize a collection event.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Custom parameters can be set with the keyword arguments,
        they will be passed to the GemEquipmentHandlers callback
        :func:`secsgem.gem.equipmenthandler.GemEquipmentHandler.on_dv_value_request`.

        If use_callbacks is disabled, you can set the value with the value property.

        Args:
            ceid: ID of the collection event
            name: long name of the collection event
            data_values: data values available for this event
            **kwargs: additional attributes for object

        """
        self.ceid = ceid if not isinstance(ceid, CollectionEventId) else ceid.value
        self.name = name
        self.data_values = data_values

        self.id_type: type[secsgem.secs.variables.Base]

        if isinstance(self.ceid, int):
            self.id_type = secsgem.secs.variables.U4
        else:
            self.id_type = secsgem.secs.variables.String

        for key, value in kwargs.items():
            setattr(self, key, value)



"""Class for stream 02 function 22."""

from secsgem.secs.data_items import CMDA
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F22(SecsStreamFunction):
    """Remote command - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F22
        CMDA: U1/I1

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F22(secsgem.secs.data_items.CMDA.DONE)
        S2F22
          <U1 0 > .

    Data Items:
        - :class:`CMDA <secsgem.secs.data_items.CMDA>`

    """

    _stream = 2
    _function = 22

    _data_format = CMDA

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""module imports."""

from .alarm import Alarm
from .collection_event import CollectionEvent, CollectionEventId
from .collection_event_link import CollectionEventLink
from .collection_event_report import CollectionEventReport
from .data_value import DataValue
from .equipment_constant import EquipmentConstant, EquipmentConstantId
from .equipmenthandler import GemEquipmentHandler
from .handler import GemHandler
from .hosthandler import GemHostHandler
from .remote_command import RemoteCommand, RemoteCommandId
from .status_variable import StatusVariable, StatusVariableId

__all__ = [
    "GemHandler", "GemEquipmentHandler", "GemHostHandler",
    "RemoteCommand", "RemoteCommandId", "Alarm", "EquipmentConstant", "EquipmentConstantId", "CollectionEventReport",
    "CollectionEventLink",   "CollectionEvent", "CollectionEventId", "StatusVariable", "StatusVariableId",
     "DataValue",
]



"""MEXP data item."""
from .. import variables
from .base import DataItemBase


class MEXP(DataItemBase):
    """Message expected.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 6

    **Used In Function**
        - :class:`SecsS09F13 <secsgem.secs.functions.SecsS09F13>`

    """

    __type__ = variables.String
    __count__ = 6



"""Header for the hsms select response."""

from .header import HsmsHeader, HsmsSType


class HsmsSelectRspHeader(HsmsHeader):
    """Header for Select Response.

    Header for message with SType 2.
    """

    def __init__(self, system: int):
        """Initialize a hsms select response.

        :param system: message ID
        :type system: integer

        Example:
            >>> import secsgem.hsms
            >>>
            >>> secsgem.hsms.HsmsSelectRspHeader(24)
            HsmsSelectRspHeader({session_id:0xffff, stream:00, function:00, p_type:0x00, s_type:0x02, \
system:0x00000018, require_response:False})
        """
        super().__init__(system, 0xFFFF, 0, 0, False, 0x00, HsmsSType.SELECT_RSP)



"""Class for stream 07 function 18."""

from secsgem.secs.data_items import ACKC7
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F18(SecsStreamFunction):
    """delete process program - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F18
        ACKC7: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F18(secsgem.secs.data_items.ACKC7.MODE_UNSUPPORTED)
        S7F18
          <B 0x5> .

    Data Items:
        - :class:`ACKC7 <secsgem.secs.data_items.ACKC7>`

    """

    _stream = 7
    _function = 18

    _data_format = ACKC7

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""EXTYPE data item."""
from .. import variables
from .base import DataItemBase


class EXTYPE(DataItemBase):
    """Exception type.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS05F09 <secsgem.secs.functions.SecsS05F09>`
        - :class:`SecsS05F11 <secsgem.secs.functions.SecsS05F11>`

    """

    __type__ = variables.String



"""Class for stream 01 function 21."""

from secsgem.secs.data_items import VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F21(SecsStreamFunction):
    """Data variable namelist request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F21
        [
            VID: U1/U2/U4/U8/I1/I2/I4/I8/A
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F21([1, "VARIABLEID"])
        S1F21 W
          <L [2]
            <U1 1 >
            <A "VARIABLEID">
          > .

    Data Items:
        - :class:`VID <secsgem.secs.data_items.VID>`

    """

    _stream = 1
    _function = 21

    _data_format = [VID]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 01 function 15."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F15(SecsStreamFunction):
    """request offline.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F15
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F15()
        S1F15 W .

    """

    _stream = 1
    _function = 15

    _data_format = None

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 02 function 45."""

from secsgem.secs.data_items import DATAID, LIMITID, LOWERDB, UPPERDB, VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F45(SecsStreamFunction):
    """Define variable limit attributes.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F45
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            DATA: [
                {
                    VID: U1/U2/U4/U8/I1/I2/I4/I8/A
                    DATA: [
                        {
                            LIMITID: B[1]
                            DATA: {
                                UPPERDB: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A[1]
                                LOWERDB: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A[1]
                            }
                        }
                        ...
                    ]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F45({
        ...     "DATAID": 1,
        ...     "DATA": [{
        ...         "VID": 2,
        ...         "DATA": [{
        ...             "LIMITID": 3,
        ...             "DATA": [30, 20]
        ...         }]
        ...     }]})
        S2F45 W
          <L [2]
            <U1 1 >
            <L [1]
              <L [2]
                <U1 2 >
                <L [1]
                  <L [2]
                    <B 0x3>
                    <L [2]
                      <U1 30 >
                      <U1 20 >
                    >
                  >
                >
              >
            >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`VID <secsgem.secs.data_items.VID>`
        - :class:`LIMITID <secsgem.secs.data_items.LIMITID>`
        - :class:`UPPERDB <secsgem.secs.data_items.UPPERDB>`
        - :class:`LOWERDB <secsgem.secs.data_items.LOWERDB>`

    """

    _stream = 2
    _function = 45

    _data_format = [
        DATAID,
        [
            [
                VID,
                [
                    [
                        LIMITID,
                        [
                            UPPERDB,
                            LOWERDB
                        ]
                    ]
                ]
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True



"""SECS 4 byte signed integer variable type."""

from .base_number import BaseNumber


class I4(BaseNumber):
    """Secs type for 4 byte signed data.

    :param value: initial value
    :type value: list/integer
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o34
    text_code = "I4"
    _base_type = int
    _min = -2147483648
    _max = 2147483647
    _bytes = 4
    _struct_code = "l"
    preferred_types = [int]



"""SECS boolean variable type."""

from .base import Base


class Boolean(Base):
    """Secs type for boolean data."""

    format_code = 0o11
    text_code = "BOOLEAN"
    preferred_types = [bool]

    _true_strings = ["TRUE", "YES"]
    _false_strings = ["FALSE", "NO"]

    def __init__(self, value=None, count=-1):
        """Initialize a boolean secs variable.

        :param value: initial value
        :type value: list/boolean
        :param count: number of items this value
        :type count: integer
        """
        super().__init__()

        self.value = []
        self.count = count
        if value is not None:
            self.set(value)

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        if len(self.value) == 0:
            return f"<{self.text_code}>"

        data = " ".join([str(value) for value in self.value])

        return f"<{self.text_code} {data} >"

    def __len__(self):
        """Get the length."""
        return len(self.value)

    def __getitem__(self, key):
        """Get an item using the indexer operator."""
        return self.value[key]

    def __setitem__(self, key, item):
        """Set an item using the indexer operator."""
        self.value[key] = item

    def __eq__(self, other):
        """Check equality with other object."""
        if isinstance(other, Base):
            if other.is_dynamic:
                return other.value.value == self.value
            return other.value == self.value

        if isinstance(other, list):
            return other == self.value

        return [other] == self.value

    def __hash__(self):
        """Get data item for hashing."""
        return hash(str(self.value))

    def _check_single_item_support(self, value):
        if isinstance(value, bool):
            return True

        if isinstance(value, int):
            if 0 <= value <= 1:
                return True
            return False

        if isinstance(value, str):
            if value.upper() in self._true_strings or value.upper() in self._false_strings:
                return True

            return False

        return False

    def supports_value(self, value) -> bool:
        """Check if the current instance supports the provided value.

        :param value: value to test
        :type value: any
        """
        if isinstance(value, (list, tuple)):
            return self._supports_value_list(value)

        if isinstance(value, bytearray):
            return self._supports_value_bytearray(value)

        return self._check_single_item_support(value)

    def _supports_value_list(self, value) -> bool:
        if 0 < self.count < len(value):
            return False

        return all(self._check_single_item_support(item) for item in value)

    def _supports_value_bytearray(self, value) -> bool:
        if 0 < self.count < len(value):
            return False

        return all(0 <= char <= 1 for char in value)

    def __convert_single_item(self, value):
        if isinstance(value, bool):
            return value

        if isinstance(value, int):
            if not 0 <= value <= 1:
                raise ValueError(f"Value {value} out of bounds")

            return bool(value)

        if isinstance(value, str):
            if value.upper() in self._true_strings:
                return True

            if value.upper() in self._false_strings:
                return False

            raise ValueError(f"Value {value} out of bounds")

        raise ValueError(f"Can't convert value {value}")

    def set(self, value):
        """Set the internal value to the provided value.

        :param value: new value
        :type value: list/boolean
        """
        if isinstance(value, (list, tuple)):
            if 0 <= self.count < len(value):
                raise ValueError(f"Value longer than {self.count} chars")

            self.value = [self.__convert_single_item(item) for item in value]
        elif isinstance(value, bytearray):
            if 0 <= self.count < len(value):
                raise ValueError(f"Value longer than {self.count} chars")

            new_value = []
            for char in value:
                if not 0 <= char <= 1:
                    raise ValueError(f"Value {char} out of bounds")

                new_value.append(char)

            self.value = new_value
        else:
            self.value = [self.__convert_single_item(value)]

    def get(self):
        """Return the internal value.

        :returns: internal value
        :rtype: list/boolean
        """
        if len(self.value) == 1:
            return self.value[0]

        return self.value

    def encode(self):
        """Encode the value to secs data.

        :returns: encoded data bytes
        :rtype: string
        """
        result = self.encode_item_header(len(self.value))

        for value in self.value:
            if value:
                result += b"\1"
            else:
                result += b"\0"

        return result

    def decode(self, data, start=0):
        """Decode the secs byte data to the value.

        :param data: encoded data bytes
        :type data: string
        :param start: start position of value the data
        :type start: integer
        :returns: new start position
        :rtype: integer
        """
        (text_pos, _, length) = self.decode_item_header(data, start)

        result = []

        for _ in range(length):
            if bytearray(data)[text_pos] == 0:
                result.append(False)
            else:
                result.append(True)

            text_pos += 1

        self.set(result)

        return text_pos



"""Class for stream 12 function 19."""

from secsgem.secs.data_items import DATLC, MAPER
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F19(SecsStreamFunction):
    """map error report - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F19
        {
            MAPER: B[1]
            DATLC: U1
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F19({"MAPER": secsgem.secs.data_items.MAPER.INVALID_DATA, "DATLC": 0})
        S12F19
          <L [2]
            <B 0x1>
            <U1 0 >
          > .

    Data Items:
        - :class:`MAPER <secsgem.secs.data_items.MAPER>`
        - :class:`DATLC <secsgem.secs.data_items.DATLC>`

    """

    _stream = 12
    _function = 19

    _data_format = [
        MAPER,
        DATLC
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""UNITS data item."""
from .. import variables
from .base import DataItemBase


class UNITS(DataItemBase):
    """Units identifier.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS01F12 <secsgem.secs.functions.SecsS01F12>`
        - :class:`SecsS01F22 <secsgem.secs.functions.SecsS01F22>`
        - :class:`SecsS02F30 <secsgem.secs.functions.SecsS02F30>`
        - :class:`SecsS02F48 <secsgem.secs.functions.SecsS02F48>`

    """

    __type__ = variables.String



"""STRID data item."""
from .. import variables
from .base import DataItemBase


class STRID(DataItemBase):
    """Stream ID.

    :Type: :class:`U1 <secsgem.secs.variables.U1>`
    :Length: 1

    **Used In Function**
        - :class:`SecsS02F43 <secsgem.secs.functions.SecsS02F43>`
        - :class:`SecsS02F44 <secsgem.secs.functions.SecsS02F44>`

    """

    __type__ = variables.U1
    __count__ = 1



"""Class for stream 06 function 20."""

from secsgem.secs.data_items import V
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F20(SecsStreamFunction):
    """individual report data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F20
        [
            V: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F20(["ASD", 1337])
        S6F20
          <L [2]
            <A "ASD">
            <U2 1337 >
          > .

    Data Items:
        - :class:`V <secsgem.secs.data_items.V>`

    """

    _stream = 6
    _function = 20

    _data_format = [V]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Class for stream 03 function 18."""

from secsgem.secs.data_items import CAACK, ERRCODE, ERRTEXT
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS03F18(SecsStreamFunction):
    """Carrier Action Acknowledge."""

    _stream = 3
    _function = 18

    _data_format = [
        CAACK,
        [
            [
                "PARAMS",
                ERRCODE,
                ERRTEXT
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Port id data item."""
from .. import variables
from .base import DataItemBase


class PTN(DataItemBase):
    """Port id number.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS03F17 <secsgem.secs.functions.SecsS03F17>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.U1,
        variables.U2,
        variables.U4,
        variables.U8,
        variables.I1,
        variables.I2,
        variables.I4,
        variables.I8,
        variables.String
    ]


"""Class for stream 02 function 35."""

from secsgem.secs.data_items import CEID, DATAID, RPTID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F35(SecsStreamFunction):
    """link event report.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F35
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            DATA: [
                {
                    CEID: U1/U2/U4/U8/I1/I2/I4/I8/A
                    RPTID: [
                        DATA: U1/U2/U4/U8/I1/I2/I4/I8/A
                        ...
                    ]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F35({"DATAID": 1, "DATA": [{"CEID": 1337, "RPTID": [1000, 1001]}]})
        S2F35 W
          <L [2]
            <U1 1 >
            <L [1]
              <L [2]
                <U2 1337 >
                <L [2]
                  <U2 1000 >
                  <U2 1001 >
                >
              >
            >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`CEID <secsgem.secs.data_items.CEID>`
        - :class:`RPTID <secsgem.secs.data_items.RPTID>`

    """

    _stream = 2
    _function = 35

    _data_format = [
        DATAID,
        [
            [
                CEID,
                [RPTID]
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True



"""Class for stream 05 function 13."""

from secsgem.secs.data_items import EXID, EXRECVRA
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F13(SecsStreamFunction):
    """exception recover - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F13
        {
            EXID: A[20]
            EXRECVRA: A[40]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F13({"EXID": "EX123", "EXRECVRA": "EXRECVRA2"})
        S5F13 W
          <L [2]
            <A "EX123">
            <A "EXRECVRA2">
          > .

    Data Items:
        - :class:`EXID <secsgem.secs.data_items.EXID>`
        - :class:`EXRECVRA <secsgem.secs.data_items.EXRECVRA>`

    """

    _stream = 5
    _function = 13

    _data_format = [
        EXID,
        EXRECVRA
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 07 function 06."""

from secsgem.secs.data_items import PPBODY, PPID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F06(SecsStreamFunction):
    """process program - data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F06
        {
            PPID: A/B[120]
            PPBODY: U1/U2/U4/U8/I1/I2/I4/I8/A/B
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F06({"PPID": "program", "PPBODY": secsgem.secs.variables.Binary("data")})
        S7F6
          <L [2]
            <A "program">
            <B 0x64 0x61 0x74 0x61>
          > .

    Data Items:
        - :class:`PPID <secsgem.secs.data_items.PPID>`
        - :class:`PPBODY <secsgem.secs.data_items.PPBODY>`

    """

    _stream = 7
    _function = 6

    _data_format = [
        PPID,
        PPBODY
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Timeout class."""
from __future__ import annotations


class _Timeout:
    def __init__(self, name: str, default: float, help_text: str) -> None:
        self._name = name
        self._default = default
        self._help = help_text

    @property
    def name(self) -> str:
        """Get the timeout name."""
        return self._name

    @property
    def default(self) -> float:
        """Get the default timeout value."""
        return self._default

    @property
    def help(self) -> str:
        """Get the help text for the timeout."""
        return self._help


class Timeouts:
    r"""Timeouts.

    Example:
        >>> import secsgem.common
        >>> timeouts = secsgem.common.Timeouts(t3=60.0)
        >>> timeouts.t3
        60.0
        >>> timeouts.t4
        10.0

    .. exec::
        import secsgem.common.timeouts

        for timeout in secsgem.common.timeouts.Timeouts._timeouts():
            print(f".. attribute:: {timeout.name}\n\n"
                    f"   :type: {timeout.default.__class__.__name__}\n"
                    f"   :value: {timeout.default}\n"
                    f"   {timeout.help}")

    """

    @classmethod
    def timeouts(cls) -> list[_Timeout]:
        """Get a list of default timeouts."""
        return [
            _Timeout("t1", 5.0, "Inter-Character Timeout"),
            _Timeout("t2", 100.0, "Protocol Timeout"),
            _Timeout("t3", 45.0, "Reply Timeout"),
            _Timeout("t4", 10.0, "Inter-Block Timeout"),
            _Timeout("t5", 10.0, "Connect Separation Time"),
            _Timeout("t6", 5.0, "Control Transaction Timeout"),
            _Timeout("t7", 8.0, "Not Selected Timeout"),
            _Timeout("t8", 5.0, "Network Intercharacter Timeout"),
        ]

    def __init__(self, **kwargs) -> None:
        """Timout initializer.

        All arguments are optional.
        The default value will be used, when an argument is omitted.

        Args:
            **kwargs: keyword arguments, see below

        Keyword Args:
            t1: Inter-Character Timeout
            t2: Protocol Timeout
            t3: Reply Timeout
            t4: Inter-Block Timeout
            t5: Connect Separation Time
            t6: Control Transaction Timeout
            t7: Not Selected Timeout
            t8: Network Intercharacter Timeout

        """
        self._data = {}

        for timeout in self.timeouts():
            self._data[timeout.name] = kwargs.get(timeout.name, timeout.default)

    def __getattr__(self, name: str):
        """Get an attribute.

        Args:
            name: attribute name

        Returns:
            attribute value

        """
        if name not in self._data:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

        return self._data[name]



"""SECS 4 byte float variable type."""

from .base_number import BaseNumber


class F4(BaseNumber):
    """Secs type for 4 byte float data.

    :param value: initial value
    :type value: list/float
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o44
    text_code = "F4"
    _base_type = float
    _min = -3.40282e+38
    _max = 3.40282e+38
    _bytes = 4
    _struct_code = "f"
    preferred_types = [float]



"""Class for stream 12 function 02."""

from secsgem.secs.data_items import SDACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F02(SecsStreamFunction):
    """map setup data - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F02
        SDACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F02(secsgem.secs.data_items.SDACK.ACK)
        S12F2
          <B 0x0> .

    Data Items:
        - :class:`SDACK <secsgem.secs.data_items.SDACK>`

    """

    _stream = 12
    _function = 2

    _data_format = SDACK

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""List of all SECS stream and functions."""
from .s00f00 import SecsS00F00
from .s01f00 import SecsS01F00
from .s01f01 import SecsS01F01
from .s01f02 import SecsS01F02
from .s01f03 import SecsS01F03
from .s01f04 import SecsS01F04
from .s01f11 import SecsS01F11
from .s01f12 import SecsS01F12
from .s01f13 import SecsS01F13
from .s01f14 import SecsS01F14
from .s01f15 import SecsS01F15
from .s01f16 import SecsS01F16
from .s01f17 import SecsS01F17
from .s01f18 import SecsS01F18
from .s01f21 import SecsS01F21
from .s01f22 import SecsS01F22
from .s01f23 import SecsS01F23
from .s01f24 import SecsS01F24
from .s02f00 import SecsS02F00
from .s02f13 import SecsS02F13
from .s02f14 import SecsS02F14
from .s02f15 import SecsS02F15
from .s02f16 import SecsS02F16
from .s02f17 import SecsS02F17
from .s02f18 import SecsS02F18
from .s02f21 import SecsS02F21
from .s02f22 import SecsS02F22
from .s02f23 import SecsS02F23
from .s02f24 import SecsS02F24
from .s02f25 import SecsS02F25
from .s02f26 import SecsS02F26
from .s02f29 import SecsS02F29
from .s02f30 import SecsS02F30
from .s02f31 import SecsS02F31
from .s02f32 import SecsS02F32
from .s02f33 import SecsS02F33
from .s02f34 import SecsS02F34
from .s02f35 import SecsS02F35
from .s02f36 import SecsS02F36
from .s02f37 import SecsS02F37
from .s02f38 import SecsS02F38
from .s02f41 import SecsS02F41
from .s02f42 import SecsS02F42
from .s02f43 import SecsS02F43
from .s02f44 import SecsS02F44
from .s02f45 import SecsS02F45
from .s02f46 import SecsS02F46
from .s02f47 import SecsS02F47
from .s02f48 import SecsS02F48
from .s02f49 import SecsS02F49
from .s02f50 import SecsS02F50
from .s03f17 import SecsS03F17
from .s03f18 import SecsS03F18
from .s05f00 import SecsS05F00
from .s05f01 import SecsS05F01
from .s05f02 import SecsS05F02
from .s05f03 import SecsS05F03
from .s05f04 import SecsS05F04
from .s05f05 import SecsS05F05
from .s05f06 import SecsS05F06
from .s05f07 import SecsS05F07
from .s05f08 import SecsS05F08
from .s05f09 import SecsS05F09
from .s05f10 import SecsS05F10
from .s05f11 import SecsS05F11
from .s05f12 import SecsS05F12
from .s05f13 import SecsS05F13
from .s05f14 import SecsS05F14
from .s05f15 import SecsS05F15
from .s05f16 import SecsS05F16
from .s05f17 import SecsS05F17
from .s05f18 import SecsS05F18
from .s06f00 import SecsS06F00
from .s06f01 import SecsS06F01
from .s06f02 import SecsS06F02
from .s06f05 import SecsS06F05
from .s06f06 import SecsS06F06
from .s06f07 import SecsS06F07
from .s06f08 import SecsS06F08
from .s06f11 import SecsS06F11
from .s06f12 import SecsS06F12
from .s06f15 import SecsS06F15
from .s06f16 import SecsS06F16
from .s06f19 import SecsS06F19
from .s06f20 import SecsS06F20
from .s06f21 import SecsS06F21
from .s06f22 import SecsS06F22
from .s06f23 import SecsS06F23
from .s06f24 import SecsS06F24
from .s07f00 import SecsS07F00
from .s07f01 import SecsS07F01
from .s07f02 import SecsS07F02
from .s07f03 import SecsS07F03
from .s07f04 import SecsS07F04
from .s07f05 import SecsS07F05
from .s07f06 import SecsS07F06
from .s07f17 import SecsS07F17
from .s07f18 import SecsS07F18
from .s07f19 import SecsS07F19
from .s07f20 import SecsS07F20
from .s07f25 import SecsS07F25
from .s07f26 import SecsS07F26
from .s09f00 import SecsS09F00
from .s09f01 import SecsS09F01
from .s09f03 import SecsS09F03
from .s09f05 import SecsS09F05
from .s09f07 import SecsS09F07
from .s09f09 import SecsS09F09
from .s09f11 import SecsS09F11
from .s09f13 import SecsS09F13
from .s10f00 import SecsS10F00
from .s10f01 import SecsS10F01
from .s10f02 import SecsS10F02
from .s10f03 import SecsS10F03
from .s10f04 import SecsS10F04
from .s12f00 import SecsS12F00
from .s12f01 import SecsS12F01
from .s12f02 import SecsS12F02
from .s12f03 import SecsS12F03
from .s12f04 import SecsS12F04
from .s12f05 import SecsS12F05
from .s12f06 import SecsS12F06
from .s12f07 import SecsS12F07
from .s12f08 import SecsS12F08
from .s12f09 import SecsS12F09
from .s12f10 import SecsS12F10
from .s12f11 import SecsS12F11
from .s12f12 import SecsS12F12
from .s12f13 import SecsS12F13
from .s12f14 import SecsS12F14
from .s12f15 import SecsS12F15
from .s12f16 import SecsS12F16
from .s12f17 import SecsS12F17
from .s12f18 import SecsS12F18
from .s12f19 import SecsS12F19
from .s14f00 import SecsS14F00
from .s14f01 import SecsS14F01
from .s14f02 import SecsS14F02
from .s14f03 import SecsS14F03
from .s14f04 import SecsS14F04

secs_streams_functions = [
    SecsS00F00,
    SecsS01F00,
    SecsS01F01,
    SecsS01F02,
    SecsS01F03,
    SecsS01F04,
    SecsS01F11,
    SecsS01F12,
    SecsS01F13,
    SecsS01F14,
    SecsS01F15,
    SecsS01F16,
    SecsS01F17,
    SecsS01F18,
    SecsS01F21,
    SecsS01F22,
    SecsS01F23,
    SecsS01F24,
    SecsS02F00,
    SecsS02F13,
    SecsS02F14,
    SecsS02F15,
    SecsS02F16,
    SecsS02F17,
    SecsS02F18,
    SecsS02F21,
    SecsS02F22,
    SecsS02F23,
    SecsS02F24,
    SecsS02F25,
    SecsS02F26,
    SecsS02F29,
    SecsS02F30,
    SecsS02F31,
    SecsS02F32,
    SecsS02F33,
    SecsS02F34,
    SecsS02F35,
    SecsS02F36,
    SecsS02F37,
    SecsS02F38,
    SecsS02F41,
    SecsS02F42,
    SecsS02F43,
    SecsS02F44,
    SecsS02F45,
    SecsS02F46,
    SecsS02F47,
    SecsS02F48,
    SecsS02F49,
    SecsS02F50,
    SecsS03F17,
    SecsS03F18,
    SecsS05F00,
    SecsS05F01,
    SecsS05F02,
    SecsS05F03,
    SecsS05F04,
    SecsS05F05,
    SecsS05F06,
    SecsS05F07,
    SecsS05F08,
    SecsS05F09,
    SecsS05F10,
    SecsS05F11,
    SecsS05F12,
    SecsS05F13,
    SecsS05F14,
    SecsS05F15,
    SecsS05F16,
    SecsS05F17,
    SecsS05F18,
    SecsS06F00,
    SecsS06F01,
    SecsS06F02,
    SecsS06F05,
    SecsS06F06,
    SecsS06F07,
    SecsS06F08,
    SecsS06F11,
    SecsS06F12,
    SecsS06F15,
    SecsS06F16,
    SecsS06F19,
    SecsS06F20,
    SecsS06F21,
    SecsS06F22,
    SecsS06F23,
    SecsS06F24,
    SecsS07F00,
    SecsS07F01,
    SecsS07F02,
    SecsS07F03,
    SecsS07F04,
    SecsS07F05,
    SecsS07F06,
    SecsS07F17,
    SecsS07F18,
    SecsS07F19,
    SecsS07F20,
    SecsS07F25,
    SecsS07F26,
    SecsS09F00,
    SecsS09F01,
    SecsS09F03,
    SecsS09F05,
    SecsS09F07,
    SecsS09F09,
    SecsS09F11,
    SecsS09F13,
    SecsS10F00,
    SecsS10F01,
    SecsS10F02,
    SecsS10F03,
    SecsS10F04,
    SecsS12F00,
    SecsS12F01,
    SecsS12F02,
    SecsS12F03,
    SecsS12F04,
    SecsS12F05,
    SecsS12F06,
    SecsS12F07,
    SecsS12F08,
    SecsS12F09,
    SecsS12F10,
    SecsS12F11,
    SecsS12F12,
    SecsS12F13,
    SecsS12F14,
    SecsS12F15,
    SecsS12F16,
    SecsS12F17,
    SecsS12F18,
    SecsS12F19,
    SecsS14F00,
    SecsS14F01,
    SecsS14F02,
    SecsS14F03,
    SecsS14F04,
]



"""Class for stream 12 function 13."""

from secsgem.secs.data_items import IDTYP, MID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F13(SecsStreamFunction):
    """map data type 1 - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F13
        {
            MID: A/B[80]
            IDTYP: B[1]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F13({"MID": "materialID", "IDTYP": secsgem.secs.data_items.IDTYP.WAFER})
        S12F13 W
          <L [2]
            <A "materialID">
            <B 0x0>
          > .

    Data Items:
        - :class:`MID <secsgem.secs.data_items.MID>`
        - :class:`IDTYP <secsgem.secs.data_items.IDTYP>`

    """

    _stream = 12
    _function = 13

    _data_format = [
        MID,
        IDTYP
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 12 function 01."""

from secsgem.secs.data_items import (
    COLCT,
    DUTMS,
    FFROT,
    FNLOC,
    IDTYP,
    MID,
    NULBC,
    ORLOC,
    PRAXI,
    PRDCT,
    REFP,
    ROWCT,
    RPSEL,
    XDIES,
    YDIES,
)
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F01(SecsStreamFunction):
    """map setup data - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F01
        {
            MID: A/B[80]
            IDTYP: B[1]
            FNLOC: U2
            FFROT: U2
            ORLOC: B[1]
            RPSEL: U1
            REFP: [
                DATA: I1/I2/I4/I8
                ...
            ]
            DUTMS: A
            XDIES: U1/U2/U4/U8/F4/F8
            YDIES: U1/U2/U4/U8/F4/F8
            ROWCT: U1/U2/U4/U8
            COLCT: U1/U2/U4/U8
            NULBC: U1/A
            PRDCT: U1/U2/U4/U8
            PRAXI: B[1]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F01({
        ...     "MID": "materialID",
        ...     "IDTYP": secsgem.secs.data_items.IDTYP.WAFER,
        ...     "FNLOC": 0,
        ...     "FFROT": 0,
        ...     "ORLOC": secsgem.secs.data_items.ORLOC.UPPER_LEFT,
        ...     "RPSEL": 0,
        ...     "REFP": [[1,2], [2,3]],
        ...     "DUTMS": "unit",
        ...     "XDIES": 100,
        ...     "YDIES": 100,
        ...     "ROWCT": 10,
        ...     "COLCT": 10,
        ...     "NULBC": "{x}",
        ...     "PRDCT": 100,
        ...     "PRAXI": secsgem.secs.data_items.PRAXI.ROWS_TOP_INCR})
        S12F1 W
          <L [15]
            <A "materialID">
            <B 0x0>
            <U2 0 >
            <U2 0 >
            <B 0x2>
            <U1 0 >
            <L [2]
              <I1 1 2 >
              <I1 2 3 >
            >
            <A "unit">
            <U1 100 >
            <U1 100 >
            <U1 10 >
            <U1 10 >
            <A "{x}">
            <U1 100 >
            <B 0x0>
          > .

    Data Items:
        - :class:`MID <secsgem.secs.data_items.MID>`
        - :class:`IDTYP <secsgem.secs.data_items.IDTYP>`
        - :class:`FNLOC <secsgem.secs.data_items.FNLOC>`
        - :class:`FFROT <secsgem.secs.data_items.FFROT>`
        - :class:`ORLOC <secsgem.secs.data_items.ORLOC>`
        - :class:`RPSEL <secsgem.secs.data_items.RPSEL>`
        - :class:`REFP <secsgem.secs.data_items.REFP>`
        - :class:`DUTMS <secsgem.secs.data_items.DUTMS>`
        - :class:`XDIES <secsgem.secs.data_items.XDIES>`
        - :class:`YDIES <secsgem.secs.data_items.YDIES>`
        - :class:`ROWCT <secsgem.secs.data_items.ROWCT>`
        - :class:`COLCT <secsgem.secs.data_items.COLCT>`
        - :class:`NULBC <secsgem.secs.data_items.NULBC>`
        - :class:`PRDCT <secsgem.secs.data_items.PRDCT>`
        - :class:`PRAXI <secsgem.secs.data_items.PRAXI>`

    """

    _stream = 12
    _function = 1

    _data_format = [
        MID,
        IDTYP,
        FNLOC,
        FFROT,
        ORLOC,
        RPSEL,
        [REFP],
        DUTMS,
        XDIES,
        YDIES,
        ROWCT,
        COLCT,
        NULBC,
        PRDCT,
        PRAXI
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Contains helper functions."""

from .block_send_info import BlockSendInfo
from .byte_queue import ByteQueue
from .callbacks import CallbackHandler
from .connection import Connection
from .events import EventProducer
from .header import Header
from .helpers import format_hex, function_name, indent_block, is_errorcode_ewouldblock, is_windows
from .message import Block, Message
from .protocol import Protocol
from .protocol_dispatcher import ProtocolDispatcher
from .settings import DeviceType, Settings
from .state_machine import State, StateMachine, Transition, UnknownTransitionError, WrongSourceStateError
from .tcp_client_connection import TcpClientConnection
from .tcp_server_connection import TcpServerConnection
from .timeouts import Timeouts

__all__ = [
    "BlockSendInfo",
    "ByteQueue",
    "CallbackHandler",
    "Connection",
    "EventProducer",
    "Header",
    "format_hex", "function_name", "indent_block", "is_windows", "is_errorcode_ewouldblock",
    "Message", "Block",
    "Protocol",
    "ProtocolDispatcher",
    "Settings", "DeviceType",
    "State", "StateMachine", "Transition", "UnknownTransitionError", "WrongSourceStateError",
    "TcpClientConnection",
    "TcpServerConnection",
    "Timeouts",
]



"""Class for stream 05 function 01."""

from secsgem.secs.data_items import ALCD, ALID, ALTX
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F01(SecsStreamFunction):
    """alarm report - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F01
        {
            ALCD: B[1]
            ALID: U1/U2/U4/U8/I1/I2/I4/I8
            ALTX: A[120]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F01({
        ...     "ALCD": secsgem.secs.data_items.ALCD.PERSONAL_SAFETY |
        ...             secsgem.secs.data_items.ALCD.ALARM_SET,
        ...     "ALID": 100,
        ...     "ALTX": "text"})
        S5F1
          <L [3]
            <B 0x81>
            <U1 100 >
            <A "text">
          > .

    Data Items:
        - :class:`ALCD <secsgem.secs.data_items.ALCD>`
        - :class:`ALID <secsgem.secs.data_items.ALID>`
        - :class:`ALTX <secsgem.secs.data_items.ALTX>`

    """

    _stream = 5
    _function = 1

    _data_format = [
        ALCD,
        ALID,
        ALTX
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 01 function 24."""

from secsgem.secs.data_items import CEID, CENAME, VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F24(SecsStreamFunction):
    """Collection event namelist.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F24
        [
            {
                CEID: U1/U2/U4/U8/I1/I2/I4/I8/A
                CENAME: A
                VID: [
                    DATA: U1/U2/U4/U8/I1/I2/I4/I8/A
                    ...
                ]
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F24([{"CEID": 1, "CENAME": "CE1", "VID": [1, "VARIABLEID"]}])
        S1F24
          <L [1]
            <L [3]
              <U1 1 >
              <A "CE1">
              <L [2]
                <U1 1 >
                <A "VARIABLEID">
              >
            >
          > .

    Data Items:
        - :class:`CEID <secsgem.secs.data_items.CEID>`
        - :class:`CENAME <secsgem.secs.data_items.CENAME>`
        - :class:`VID <secsgem.secs.data_items.VID>`

    """

    _stream = 1
    _function = 24

    _data_format = [
        [
            CEID,
            CENAME,
            [VID]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 06 function 05."""

from secsgem.secs.data_items import DATAID, DATALENGTH
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F05(SecsStreamFunction):
    """multi block data inquiry.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F05
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            DATALENGTH: U1/U2/U4/U8/I1/I2/I4/I8
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F05({"DATAID": 1, "DATALENGTH": 1337})
        S6F5 W
          <L [2]
            <U1 1 >
            <U2 1337 >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`DATALENGTH <secsgem.secs.data_items.DATALENGTH>`

    """

    _stream = 6
    _function = 5

    _data_format = [
        DATAID,
        DATALENGTH
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Event Notification (collection events) capability."""
from __future__ import annotations

import threading

import secsgem.common
import secsgem.secs

from .capability import Capability
from .collection_event import CollectionEvent, CollectionEventId
from .collection_event_link import CollectionEventLink
from .collection_event_report import CollectionEventReport
from .handler import GemHandler


class CollectionEventCapability(GemHandler, Capability):
    """Event Notification (collection events) capability on GEM equipment."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self._collection_events: dict[int | str | CollectionEventId, CollectionEvent] = {
            CollectionEventId.EQUIPMENT_OFFLINE.value: CollectionEvent(
                CollectionEventId.EQUIPMENT_OFFLINE,
                "EquipmentOffline",
                []),
            CollectionEventId.CONTROL_STATE_LOCAL.value: CollectionEvent(
                CollectionEventId.CONTROL_STATE_LOCAL,
                "ControlStateLocal",
                []),
            CollectionEventId.CONTROL_STATE_REMOTE.value: CollectionEvent(
                CollectionEventId.CONTROL_STATE_REMOTE,
                "ControlStateRemote",
                []),
            CollectionEventId.CMD_START_DONE.value: CollectionEvent(
                CollectionEventId.CMD_START_DONE,
                "CmdStartDone",
                []),
            CollectionEventId.CMD_STOP_DONE.value: CollectionEvent(
                CollectionEventId.CMD_STOP_DONE,
                "CmdStopDone",
                []),
        }

        self._registered_reports: dict[int | str, CollectionEventReport] = {}
        self._registered_collection_events: dict[int | str, CollectionEventLink] = {}

    @property
    def collection_events(self) -> dict[int | str | CollectionEventId, CollectionEvent]:
        """Get list of the collection events.

        Returns:
            Collection event list

        """
        return self._collection_events

    @property
    def registered_reports(self) -> dict[int | str, CollectionEventReport]:
        """Get list of the subscribed reports.

        Returns:
            Collection event report list

        """
        return self._registered_reports

    @property
    def registered_collection_events(self) -> dict[int | str, CollectionEventLink]:
        """Get list of the subscribed collection events.

        Returns:
            Collection event list

        """
        return self._registered_collection_events

    def _get_events_enabled(self) -> list[int | str]:
        """List of the enabled collection events.

        :returns: collection event
        :rtype: list of various
        """
        enabled_ceid = []

        for ceid, collection_event in self._registered_collection_events.items():
            if collection_event.enabled:
                enabled_ceid.append(ceid)

        return enabled_ceid

    def trigger_collection_events(self, ceids: list[int | str | CollectionEventId]):
        """Triggers the supplied collection events.

        Args:
            ceids: List of collection events

        """
        def _ce_sender():
            nonlocal ceids
            if not isinstance(ceids, list):
                ceids = [ceids]

            for ceid in ceids:
                if isinstance(ceid, CollectionEventId):
                    ceid = ceid.value

                if ceid in self._registered_collection_events and self._registered_collection_events[ceid].enabled:
                    reports = self._build_collection_event(ceid)

                    self.send_and_waitfor_response(self.stream_function(6, 11)(
                        {"DATAID": 1, "CEID": ceid, "RPT": reports}))

        threading.Thread(target=_ce_sender, daemon=True).start()

    def _on_s02f33(self,  # noqa: C901, pylint: disable=too-many-branches
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 2, Function 33, Define Report.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        drack = secsgem.secs.data_items.DRACK.ACK

        # pre check message for errors
        for report in function.DATA:
            if report.RPTID in self._registered_reports and len(report.VID) > 0:
                drack = secsgem.secs.data_items.DRACK.RPTID_REDEFINED
            else:
                for vid in report.VID:
                    if (vid not in self._data_values) and (vid not in self._status_variables):
                        drack = secsgem.secs.data_items.DRACK.VID_UNKNOWN

        result = self.stream_function(2, 34)(drack)

        if drack != 0:
            return result

        # no data -> remove all reports and links
        if not function.DATA:
            self._registered_collection_events.clear()
            self._registered_reports.clear()

            return result

        for report in function.DATA:
            # no vids -> remove this reports and links
            if not report.VID:
                # remove report from linked collection events
                for collection_event in list(self._registered_collection_events):
                    if report.RPTID in self._registered_collection_events[collection_event].reports:
                        self._registered_collection_events[collection_event].reports.remove(report.RPTID)
                        # remove collection event link if no collection events present
                        if not self._registered_collection_events[collection_event].reports:
                            del self._registered_collection_events[collection_event]
                # remove report
                if report.RPTID in self._registered_reports:
                    del self._registered_reports[report.RPTID]
            else:
                # add report
                self._registered_reports[report.RPTID] = CollectionEventReport(report.RPTID, report.VID)

        return result

    def _on_s02f35(self,  # noqa: C901, pylint: disable=too-many-branches
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 2, Function 35, Link event report.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        lrack = secsgem.secs.data_items.LRACK.ACK

        # pre check message for errors
        for event in function.DATA:
            if event.CEID.get() not in self._collection_events:
                lrack = secsgem.secs.data_items.LRACK.CEID_UNKNOWN
            for rptid in event.RPTID:
                if event.CEID.get() in self._registered_collection_events:
                    collection_event = self._registered_collection_events[event.CEID.get()]
                    if rptid.get() in collection_event.reports:
                        lrack = secsgem.secs.data_items.LRACK.CEID_LINKED
                if rptid.get() not in self._registered_reports:
                    lrack = secsgem.secs.data_items.LRACK.RPTID_UNKNOWN

        # pre check okay
        if lrack == 0:
            for event in function.DATA:
                # no report ids, remove all links for collection event
                if not event.RPTID:
                    if event.CEID.get() in self._registered_collection_events:
                        del self._registered_collection_events[event.CEID.get()]
                else:
                    if event.CEID.get() in self._registered_collection_events:
                        collection_event = self._registered_collection_events[event.CEID.get()]
                        for rptid in event.RPTID.get():
                            collection_event.reports.append(rptid)
                    else:
                        self._registered_collection_events[event.CEID.get()] = \
                            CollectionEventLink(self._collection_events[event.CEID.get()], event.RPTID.get())

        return self.stream_function(2, 36)(lrack)

    def _on_s02f37(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Callback handler for Stream 2, Function 37, En-/Disable Event Report.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        erack = secsgem.secs.data_items.ERACK.ACCEPTED

        if not self._set_ce_state(function.CEED.get(), function.CEID.get()):
            erack = secsgem.secs.data_items.ERACK.CEID_UNKNOWN

        return self.stream_function(2, 38)(erack)

    def _on_s06f15(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Callback handler for Stream 6, Function 15, event report request.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        ceid = function.get()

        reports = []

        if ceid in self._registered_collection_events and self._registered_collection_events[ceid].enabled:
            reports = self._build_collection_event(ceid)

        return self.stream_function(6, 16)({"DATAID": 1, "CEID": ceid, "RPT": reports})

    def _set_ce_state(self, ceed: bool, ceids: list[int | str]) -> bool:
        """En-/Disable event reports for the supplied ceids (or all, if ceid is an empty list).

        Args:
            ceed: Enable (True) or disable (False) event reports
            ceids: List of collection events

        Returns:
            True if all ceids were ok, False if illegal ceid was supplied

        """
        result = True
        if not ceids:
            for collection_event in self._registered_collection_events.values():
                collection_event.enabled = ceed
        else:
            for ceid in ceids:
                if ceid in self._registered_collection_events:
                    self._registered_collection_events[ceid].enabled = ceed
                else:
                    result = False

        return result

    def _build_collection_event(self, ceid: int | str):
        """Build reports for a collection event.

        Args:
            ceid: collection event to build

        Returns:
            collection event data

        """
        reports = []

        for rptid in self._registered_collection_events[ceid].reports:
            report = self._registered_reports[rptid]
            variables = []
            for var in report.vars:
                if var in self._status_variables:
                    value = self._get_sv_value(self._status_variables[var])
                    variables.append(value)
                elif var in self._data_values:
                    value = self._get_dv_value(self._data_values[var])
                    variables.append(value)

            reports.append({"RPTID": rptid, "V": variables})

        return reports

    def get_ceid_name(self, ceid: int | str) -> str:
        """Get the name of a collection event.

        Args:
            ceid: ID of collection event

        Returns:
            Name of the event or empty string if not found

        """
        if ceid in self._collection_events:
            return self._collection_events[ceid].name

        return ""



"""Message base class."""
from __future__ import annotations

import abc
import struct
import typing

if typing.TYPE_CHECKING:
    from .header import Header

BlockHeaderT = typing.TypeVar("BlockHeaderT", bound="Header")
BlockT = typing.TypeVar("BlockT", bound="Block")

MessageT = typing.TypeVar("MessageT", bound="Message")


class Block(abc.ABC, typing.Generic[BlockHeaderT]):
    """Base class for data block."""

    header_type: type[BlockHeaderT]
    length_format: str
    checksum_format: str

    def __init__(self, header: BlockHeaderT, data: bytes):
        """Initialize a block header.

        Args:
            header: block header
            data: block data

        """
        self._header = header
        self._data = data

    @property
    def header(self) -> BlockHeaderT:
        """Get the header."""
        return self._header

    @property
    def data(self) -> bytes:
        """Get the data."""
        return self._data

    @property
    def checksum(self) -> int:
        """Get the checksum."""
        if self.checksum_format == "":
            return 0

        calculated_checksum = 0

        for data_byte in self.header.encode() + self.data:
            calculated_checksum += data_byte

        return calculated_checksum

    def encode(self) -> bytes:
        """Encode block data.

        Returns:
            byte-encoded block

        """
        data_length = len(self.data)

        struct_args: tuple = (
            self.header.length + data_length,
            self.header.encode(),
            self.data,
        )

        if self.checksum_format != "":
            struct_args += (self.checksum, )

        return struct.pack(
            f">{self.length_format}{self.header.length}s{data_length}s{self.checksum_format}",
            *struct_args,
        )

    @classmethod
    def decode(cls: type[BlockT], data: bytes) -> BlockT | None:
        """Decode byte array hsms packet to HsmsPacket object.

        Args:
            data: byte-encode packet data

        Returns:
            received packet object

        """
        data_length = struct.unpack_from(f">{cls.length_format}", data)[0] - cls.header_type.length

        data_fields = struct.unpack(
            f">{cls.length_format}{cls.header_type.length}s{data_length}s{cls.checksum_format}",
            data,
        )

        header = cls.header_type.decode(data_fields[1])

        obj = cls(header, data_fields[2])

        if cls.checksum_format != "" and obj.checksum != data_fields[3]:
            return None

        return obj


class Message(abc.ABC, typing.Generic[BlockT]):
    """Abstract base class for a message."""

    block_size = -1
    block_type: type[BlockT]

    def __init__(self, header: BlockHeaderT, data: bytes):
        """Initialize a Message object.

        Args:
            header: header used for this message
            data: data part used for streams and functions (SType 0)

        """
        self._blocks: list[BlockT] = self._split_blocks(data, header)

    @classmethod
    def _split_blocks(cls, data: bytes, header: BlockHeaderT) -> list[BlockT]:
        if cls.block_size == -1:
            return [cls.block_type(header, data)]

        data_blocks = [data[i: i + cls.block_size] for i in range(0, len(data), cls.block_size)]

        blocks = []
        for index, block_data in enumerate(data_blocks):
            header_data = {
                "block": index + 1,
                "last_block": (index + 1) == len(data_blocks),
            }
            block_header = header.updated_with(**header_data)
            blocks.append(cls.block_type(block_header, block_data))

        return blocks

    @classmethod
    def from_block(cls: type[MessageT], block: Block) -> MessageT:
        """Initialize Message object from Block object.

        Args:
            block: block to initialize from

        Returns:
            Message object

        """
        return cls(block.header, block.data)

    @property
    @abc.abstractmethod
    def header(self) -> Header:
        """Get the header."""
        raise NotImplementedError("Message.header missing implementation")

    @property
    @abc.abstractmethod
    def data(self) -> bytes:
        """Get the header."""
        raise NotImplementedError("Message.data missing implementation")

    @property
    @abc.abstractmethod
    def complete(self) -> bool:
        """Check if the message is complete."""
        raise NotImplementedError("Message.complete missing implementation")

    @property
    def blocks(self) -> list[BlockT]:
        """Get the blocks."""
        return self._blocks

    def __str__(self) -> str:
        """Generate string representation for an object of this class."""
        return f"'header': {self.header} "

    def __repr__(self) -> str:
        """Generate textual representation for an object of this class."""
        return (
            f"{self.__class__.__name__}"
            f"({{'header': {self.header.__repr__()}, 'data': '{self.data.decode('utf-8')}'}})"
        )



"""HSMS Header for streams/functions."""

from .header import HsmsHeader, HsmsSType


class HsmsStreamFunctionHeader(HsmsHeader):
    """Header for SECS message.

    Header for message with SType 0.
    """

    def __init__(  # pylint: disable=too-many-arguments
            self,
            system: int,
            stream: int,
            function: int,
            require_response: bool,
            session_id: int):
        """Initialize a stream function secs header.

        Args:
            system: message ID
            stream: messages stream
            function: messages function
            require_response: is response expected from remote
            session_id: device / session ID

        Example:
            >>> import secsgem.hsms
            >>>
            >>> secsgem.hsms.HsmsStreamFunctionHeader(22, 1, 1, True, 100)
            HsmsStreamFunctionHeader({session_id:0x0064, stream:01, function:01, p_type:0x00, s_type:0x00, \
system:0x00000016, require_response:True})
        """
        super().__init__(system, session_id, stream, function, require_response, 0x00, HsmsSType.DATA_MESSAGE)



"""Class for stream 02 function 47."""

from secsgem.secs.data_items import VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F47(SecsStreamFunction):
    """Variable limit attribute request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F47
        [
            VID: U1/U2/U4/U8/I1/I2/I4/I8/A
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F47([1, "VARIABLEID"])
        S2F47 W
          <L [2]
            <U1 1 >
            <A "VARIABLEID">
          > .

    Data Items:
        - :class:`VID <secsgem.secs.data_items.VID>`

    """

    _stream = 2
    _function = 47

    _data_format = [VID]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""SECS string text variable type."""

import unicodedata

from .base_text import BaseText


class String(BaseText):
    """Secs type for string data.

    :param value: initial value
    :type value: string
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o20
    text_code = "A"
    preferred_types = [bytes, str]
    control_chars = "".join(chr(ch) for ch in range(256) if unicodedata.category(chr(ch))[0] == "C" or ch > 127)
    coding = "latin-1"



"""Class for stream 07 function 19."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F19(SecsStreamFunction):
    """current equipment process program - request.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F19
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F19()
        S7F19 W .

    """

    _stream = 7
    _function = 19

    _data_format = None

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 12 function 12."""

from secsgem.secs.data_items import MDACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F12(SecsStreamFunction):
    """map data type 3 - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F12
        MDACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F12(secsgem.secs.data_items.MDACK.FORMAT_ERROR)
        S12F12
          <B 0x1> .

    Data Items:
        - :class:`MDACK <secsgem.secs.data_items.MDACK>`

    """

    _stream = 12
    _function = 12

    _data_format = MDACK

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 02 function 24."""

from secsgem.secs.data_items import TIAACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F24(SecsStreamFunction):
    """Trace initialize - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F24
        TIAACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F24(secsgem.secs.data_items.TIAACK.SVID_EXCEEDED)
        S2F24
          <B 0x1> .

    Data Items:
        - :class:`TIAACK <secsgem.secs.data_items.TIAACK>`

    """

    _stream = 2
    _function = 24

    _data_format = TIAACK

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False
